"""
Module for defining how satellite data will be collected from AWS and where it will be saved.
"""

import datetime as dt
import os
import warnings
from abc import ABCMeta, abstractmethod
from typing import Any, List, Optional, Tuple, Union

from ..api.opensearch import get_tile_info, get_tile_info_id
from ..config import SHConfig
from ..constants import MimeType
from ..data_collections import DataCollection
from ..download import DownloadRequest
from ..exceptions import AwsDownloadFailedException, SHUserWarning
from ..time_utils import parse_time
from .client import AwsDownloadClient
from .constants import AwsConstants, EsaSafeType

MAX_SUPPORTED_BASELINES = {
    DataCollection.SENTINEL2_L1C: "04.00",
    DataCollection.SENTINEL2_L2A: "04.00",
}


class AwsData(metaclass=ABCMeta):
    """A base class for collecting satellite data from AWS."""

    def __init__(
        self,
        parent_folder: str = "",
        bands: Union[None, str, List[str]] = None,
        metafiles: Union[None, str, List[str]] = None,
        config: Optional[SHConfig] = None,
    ):
        """
        :param parent_folder: Folder where the fetched data will be saved.
        :param bands: List of Sentinel-2 bands for request. If parameter is set to `None` all bands will be used.
        :param metafiles: List of additional metafiles available on AWS
                          (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                          If parameter is set to `None` the list will be set automatically.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.parent_folder = parent_folder
        self.bands = self._parse_bands(bands)
        self.metafiles = self._parse_metafiles(metafiles)
        self.config = config or SHConfig()

        self.download_list: List[DownloadRequest] = []
        self.folder_list: List[str] = []

        self.base_url = self.get_base_url()
        self.base_http_url = self.get_base_url(force_http=True)

        # These need to be set by the child classes
        self.baseline: str
        self.safe_type: EsaSafeType
        self.data_collection: DataCollection
        self.date: dt.date
        self.product_id: str

    @abstractmethod
    def get_requests(self) -> Tuple[List[DownloadRequest], List[str]]:
        """Abstract class for joining together download requests"""
        raise NotImplementedError

    def _parse_bands(self, band_input: Union[None, str, List[str]]) -> List[str]:
        """
        Parses class input and verifies band names.

        :param band_input: input parameter `bands`
        :return: verified list of bands
        """
        all_bands = (
            AwsConstants.S2_L1C_BANDS
            if self.data_collection is DataCollection.SENTINEL2_L1C
            else AwsConstants.S2_L2A_BANDS
        )

        if band_input is None:
            return all_bands
        if isinstance(band_input, str):
            band_list = band_input.split(",")
        elif isinstance(band_input, list):
            band_list = band_input.copy()
        else:
            raise ValueError("bands parameter must be a list or a string")
        band_list = [band.strip().split(".")[0] for band in band_list]
        band_list = [band for band in band_list if band != ""]
        if not set(band_list) <= set(all_bands):
            raise ValueError(f"bands {band_list} must be a subset of {all_bands}")
        return band_list

    def _parse_metafiles(self, metafile_input: Union[None, str, List[str]]) -> List[str]:
        """Parses class input and verifies metadata file names.

        :param metafile_input: class input parameter `metafiles`
        :return: verified list of metadata files
        """
        all_metafiles = (
            AwsConstants.S2_L1C_METAFILES
            if self.data_collection is DataCollection.SENTINEL2_L1C
            else AwsConstants.S2_L2A_METAFILES
        )

        if metafile_input is None:
            if self.__class__.__name__ == "SafeProduct":
                return all_metafiles
            if self.__class__.__name__ == "SafeTile":
                return [metafile for metafile in all_metafiles if metafile in AwsConstants.TILE_FILES]
            return []
        if isinstance(metafile_input, str):
            metafile_list = metafile_input.split(",")
        elif isinstance(metafile_input, list):
            metafile_list = metafile_input.copy()
        else:
            raise ValueError("metafiles parameter must be a list or a string")
        metafile_list = [metafile.strip().split(".")[0] for metafile in metafile_list]
        metafile_list = [metafile for metafile in metafile_list if metafile != ""]
        if not set(metafile_list) <= set(all_metafiles):
            raise ValueError(f"metadata files {metafile_list} must be a subset of {all_metafiles}")
        return metafile_list

    def get_base_url(self, force_http: bool = False) -> str:
        """Creates base URL path

        :param force_http: `True` if HTTP base URL should be used and `False` otherwise
        :return: base url string
        """
        base_url = self.config.aws_metadata_url if force_http else "s3://"
        aws_bucket = (
            self.config.aws_s3_l1c_bucket
            if self.data_collection is DataCollection.SENTINEL2_L1C
            else self.config.aws_s3_l2a_bucket
        )

        return base_url + ("" if base_url.endswith("/") else "/") + aws_bucket

    def get_safe_type(self) -> EsaSafeType:
        """Determines the type of ESA product.

        In 2016 ESA changed structure and naming of data. Therefore, the class must
        distinguish between old product type and compact (new) product type.

        :return: type of ESA product
        :raises: ValueError
        """
        product_type = self.product_id.split("_")[1]
        if product_type.startswith("MSI"):
            return EsaSafeType.COMPACT_TYPE
        if product_type in ["OPER", "USER"]:
            return EsaSafeType.OLD_TYPE
        raise ValueError(f"Unrecognized product type of product id {self.product_id}")

    def get_baseline(self) -> str:
        """Determines the baseline number (i.e. version) of ESA .SAFE product

        :return: baseline number
        :raises: ValueError
        """
        if self.get_safe_type() is EsaSafeType.COMPACT_TYPE:
            baseline = self.product_id.split("_")[3].lstrip("N")
            if len(baseline) != 4:
                raise ValueError(f"Unable to recognize baseline number from the product id {self.product_id}")
            baseline = f"{baseline[:2]}.{baseline[2:]}"

            if baseline > MAX_SUPPORTED_BASELINES[self.data_collection]:
                message = (
                    f"Products with baseline {baseline} are not officially supported in sentinelhub-py. If you notice "
                    "any errors in naming structure of downloaded data please report an issue at "
                    "https://github.com/sentinel-hub/sentinelhub-py/issues. Pull requests are also very appreciated"
                )
                warnings.warn(message, category=SHUserWarning)

            return baseline
        return self._read_baseline_from_info()

    def _read_baseline_from_info(self) -> str:
        """Tries to find and return baseline number from either tileInfo or productInfo file.

        :return: Baseline ID
        :raises: ValueError
        """
        if hasattr(self, "tile_info"):
            return self.tile_info["datastrip"]["id"][-5:]  # type: ignore[attr-defined]
        if hasattr(self, "product_info"):
            return self.product_info["datastrips"][0]["id"][-5:]  # type: ignore[attr-defined]
        raise ValueError("No info file has been obtained yet.")

    @staticmethod
    def url_to_tile(url: str) -> Tuple[str, str, int]:
        """Extracts tile name, date and AWS index from tile url on AWS.

        :param url: class input parameter 'metafiles'
        :return: Name of tile, date and AWS index which uniquely identifies tile on AWS
        """
        info = url.strip("/").split("/")
        name = "".join(info[-7:-4])
        date = "-".join(info[-4:-1])
        return name, date, int(info[-1])

    def sort_download_list(self) -> None:
        """Method for sorting the list of download requests. Band images have priority before metadata files. If bands
        images or metadata files are specified with a list they will be sorted in the same order as in the list.
        Otherwise, they will be sorted alphabetically (band B8A will be between B08 and B09).
        """

        def aws_sort_function(download_request: DownloadRequest) -> Tuple[int, str, int]:
            data_name = download_request.properties["data_name"]
            if "product_name" in download_request.properties:
                product_name = download_request.properties["product_name"]
            else:
                if download_request.url is None:
                    raise ValueError(f"Faulty request {download_request}, no URL specified.")
                product_name = self._url_to_props(download_request.url)[0]
            if data_name in self.bands:
                return 0, product_name, self.bands.index(data_name)
            return 1, product_name, self.metafiles.index(data_name)

        self.download_list.sort(key=aws_sort_function)

    def structure_recursion(self, struct: dict, folder: str) -> None:
        """From nested dictionaries representing .SAFE structure it recursively extracts all the files that need to be
        downloaded and stores them into class attribute `download_list`.

        :param struct: nested dictionaries representing a part of .SAFE structure
        :param folder: name of folder where this structure will be saved
        """
        has_subfolder = False
        for name, substruct in struct.items():
            subfolder = os.path.join(folder, name)
            if not isinstance(substruct, dict):
                product_name, data_name = self._url_to_props(substruct)
                if "." in data_name:
                    data_type = MimeType(data_name.split(".", 1)[-1])
                    data_name = data_name.rsplit(".", 1)[0]
                else:
                    data_type = MimeType.RAW
                if data_name in self.bands + self.metafiles:
                    self.download_list.append(
                        DownloadRequest(
                            url=substruct,
                            filename=subfolder,
                            data_type=data_type,
                            data_name=data_name,
                            product_name=product_name,
                        )
                    )
            else:
                has_subfolder = True
                self.structure_recursion(substruct, subfolder)
        if not has_subfolder:
            self.folder_list.append(folder)

    def _url_to_props(self, url: str) -> Tuple[str, str]:
        """Converts url back to name of product/tile and name of the file

        :param url: URL location of the data
        :return: Names of product or tile and name of a file
        """
        props = (
            (url[len(self.base_url) :] if url.startswith(self.base_url) else url[len(self.base_http_url) :])
            .lstrip("/")
            .split("/")
        )
        if props[0] == "products":
            tile_props = props[:5]
            props = props[5:]
            if props[0] == "datastrip":
                props[1] = "*"
        else:
            tile_props = props[:8]
            props = props[8:]
        return "/".join(tile_props), "/".join(props)

    @staticmethod
    def add_file_extension(filename: str, data_format: Optional[MimeType] = None, remove_path: bool = False) -> str:
        """Joins filename and corresponding file extension if it has one.

        :param filename: Name of the file without extension
        :param data_format: format of file, if `None` it will be set automatically
        :param remove_path: `True` if the path in filename string should be removed
        :return: Name of the file with extension
        """
        if data_format is None:
            data_format = AwsConstants.AWS_FILES[filename]
        if remove_path:
            filename = filename.split("/")[-1]
        if filename.startswith("datastrip"):
            filename = filename.replace("*", "0")
        if data_format is MimeType.RAW:
            return filename
        return f"{filename.replace('*', '')}.{data_format.value}"

    def has_reports(self) -> bool:
        """Products created with baseline 2.06 and greater (and some products with baseline 2.05) should have quality
        report files

        :return: `True` if the product has report xml files and `False` otherwise
        """
        # Not completely sure if this condition is correct
        return self.baseline > "02.05" or (
            self.baseline == "02.05" and self.date >= dt.date(year=2017, month=10, day=12)
        )

    def is_early_compact_l2a(self) -> bool:
        """Check if product is early version of compact L2A product

        :return: `True` if product is early version of compact L2A product and `False` otherwise
        """
        return (
            self.data_collection is DataCollection.SENTINEL2_L2A
            and self.safe_type is EsaSafeType.COMPACT_TYPE
            and "00.01" < self.baseline <= "02.06"
        )


class AwsProduct(AwsData):
    """Class for collecting Sentinel-2 products data from AWS."""

    def __init__(self, product_id: str, tile_list: Union[None, str, List[str]] = None, **kwargs: Any):
        """
        :param product_id: ESA ID of the product
        :param tile_list: list of tile names
        :param parent_folder: location of the directory where the fetched data will be saved.
        :param bands: List of Sentinel-2 bands for request. If parameter is set to `None` all bands will be used.
        :param metafiles: List of additional metafiles available on AWS
                          (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                          If parameter is set to `None` the list will be set automatically.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.product_id = product_id.split(".")[0]
        self.tile_list = self.parse_tile_list(tile_list)

        self.data_collection = self.get_data_collection()
        self.safe_type = self.get_safe_type()

        super().__init__(**kwargs)

        self.date = self.get_date()
        self.product_url = self.get_product_url()

        client = AwsDownloadClient(config=self.config)
        self.product_info = client.get_json_dict(self.get_url(AwsConstants.PRODUCT_INFO))
        self.baseline = self.get_baseline()

    @staticmethod
    def parse_tile_list(tile_input: Union[None, str, List[str]]) -> Optional[List[str]]:
        """Parses class input and verifies band names.

        :param tile_input: class input parameter `tile_list`
        :return: parsed list of tiles
        """
        if tile_input is None:
            return None
        if isinstance(tile_input, str):
            tile_list = tile_input.split(",")
        elif isinstance(tile_input, list):
            tile_list = tile_input.copy()
        else:
            raise ValueError("tile_list parameter must be a list of tile names")
        tile_list = [AwsTile.parse_tile_name(tile_name) for tile_name in tile_list]
        return tile_list

    def get_requests(self) -> Tuple[List[DownloadRequest], List[str]]:
        """Creates product structure and returns list of files for download.

        :return: List of download requests and list of empty folders that need to be created
        """
        self.download_list = [
            DownloadRequest(
                url=self.get_url(metafile),
                filename=self.get_filepath(metafile),
                data_type=AwsConstants.AWS_FILES[metafile],
                data_name=metafile,
            )
            for metafile in self.metafiles
            if metafile in AwsConstants.PRODUCT_FILES
        ]

        tile_parent_folder = os.path.join(self.parent_folder, self.product_id)
        for tile_info in self.product_info["tiles"]:
            tile_name, date, aws_index = self.url_to_tile(self.get_tile_url(tile_info))
            if self.tile_list is None or AwsTile.parse_tile_name(tile_name) in self.tile_list:
                tile_downloads, tile_folders = AwsTile(
                    tile_name,
                    date,
                    aws_index,
                    parent_folder=tile_parent_folder,
                    bands=self.bands,
                    metafiles=self.metafiles,
                    data_collection=self.data_collection,
                ).get_requests()
                self.download_list.extend(tile_downloads)
                self.folder_list.extend(tile_folders)
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_data_collection(self) -> DataCollection:
        """The method determines data collection from product ID.

        :return: Data collection of the product
        :raises: ValueError
        """
        product_type = self.product_id.split("_")[1]
        if product_type.endswith("L1C") or product_type == "OPER":
            return DataCollection.SENTINEL2_L1C
        if product_type.endswith("L2A") or product_type == "USER":
            return DataCollection.SENTINEL2_L2A
        raise ValueError(f"Unknown data collection of product {self.product_id}")

    def get_date(self) -> dt.date:
        """Collects sensing date of the product.

        :return: Sensing date
        """
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = self.product_id.split("_")[-2]
            date = (name[1:5], name[5:7], name[7:9])
        else:
            name = self.product_id.split("_")[2]
            date = (name[:4], name[4:6], name[6:8])

        return dt.date(year=int(date[0]), month=int(date[1]), day=int(date[2]))

    def get_url(self, filename: str, data_format: Optional[MimeType] = None) -> str:
        """Creates url of file location on AWS.

        :param filename: name of file
        :param data_format: format of file, if `None` it will be set automatically
        :return: url of file location
        """
        product_url = self.product_url
        force_http = filename in [AwsConstants.PRODUCT_INFO, AwsConstants.METADATA]
        if product_url is None or force_http:
            product_url = self.get_product_url(force_http=force_http)
        return f"{product_url}/{self.add_file_extension(filename, data_format)}"

    def get_product_url(self, force_http: bool = False) -> str:
        """Creates base url of product location on AWS.

        :param force_http: `True` if HTTP base URL should be used and `False` otherwise
        :return: url of product location
        """
        base_url = self.base_http_url if force_http else self.base_url
        return f"{base_url}/products/{self.date.year}/{self.date.month}/{self.date.day}/{self.product_id}"

    def get_tile_url(self, tile_info: dict) -> str:
        """Collects tile url from `productInfo.json` file.

        :param tile_info: information about tile from `productInfo.json`
        :return: url of tile location
        """
        return f"{self.base_url}/{tile_info['path']}"

    def get_filepath(self, filename: str) -> str:
        """Creates file path for the file.

        :param filename: name of the file
        :return: filename with path on disk
        """
        return os.path.join(self.parent_folder, self.product_id, self.add_file_extension(filename)).replace(":", ".")


class AwsTile(AwsData):
    """Class for collecting Sentinel-2 tiles data from AWS."""

    def __init__(
        self,
        tile_name: str,
        time: str,
        aws_index: Optional[int] = None,
        data_collection: DataCollection = DataCollection.SENTINEL2_L1C,
        **kwargs: Any,
    ):
        """
        :param tile: Tile name (e.g. 'T10UEV')
        :param time: Tile sensing time in ISO8601 format
        :param aws_index: There exist Sentinel-2 tiles with the same tile and time parameter. Therefore, each tile
            on AWS also has an index which is visible in their url path. If `aws_index` is set to `None` the
            class will try to find the index automatically. If there will be multiple choices it will choose
            the lowest index and inform the user.
        :param data_collection: A collection of requested AWS data. Supported collections are Sentinel-2 L1C and
            Sentinel-2 L2A, default is Sentinel-2 L1C data.
        :param parent_folder: folder where the fetched data will be saved.
        :param bands: List of Sentinel-2 bands for request. If parameter is set to `None` all bands will be used.
        :param metafiles: List of additional metafiles available on AWS
                          (e.g. ``['metadata', 'tileInfo', 'preview/B01', 'TCI']``).
                          If parameter is set to `None` the list will be set automatically.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.tile_name = self.parse_tile_name(tile_name)

        self.timestamp: dt.date = parse_time(time, ignoretz=True)  # type: ignore
        self.date = self.timestamp.date() if isinstance(self.timestamp, dt.datetime) else self.timestamp

        self.aws_index = aws_index
        self.data_collection = data_collection

        super().__init__(**kwargs)
        self.tile_url = None

        self.aws_index = self.get_aws_index()
        self.tile_url = self.get_tile_url()
        self.tile_info = self.get_tile_info()
        if not self.tile_is_valid():
            raise ValueError("Cannot find data on AWS for specified tile, time and aws_index")

        self.product_id = self.get_product_id()
        self.safe_type = self.get_safe_type()
        self.baseline = self.get_baseline()

    @staticmethod
    def parse_tile_name(name: str) -> str:
        """
        Parses and verifies tile name.

        :param name: class input parameter `tile_name`
        :return: parsed tile name
        """
        tile_name = name.lstrip("T0")
        if len(tile_name) == 4:
            tile_name = "0" + tile_name
        if len(tile_name) != 5:
            raise ValueError(f"Invalid tile name {name}")
        return tile_name

    def get_requests(self) -> Tuple[List[DownloadRequest], List[str]]:
        """
        Creates tile structure and returns list of files for download.

        :return: List of download requests and list of empty folders that need to be created
        """
        self.download_list = []
        for data_name in [band for band in self.bands if self._band_exists(band)] + self.metafiles:
            if data_name in AwsConstants.TILE_FILES:
                url = self.get_url(data_name)
                filename = self.get_filepath(data_name)
                self.download_list.append(
                    DownloadRequest(
                        url=url, filename=filename, data_type=AwsConstants.AWS_FILES[data_name], data_name=data_name
                    )
                )
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_aws_index(self) -> int:
        """
        Returns tile index on AWS. If `tile_index` was not set during class initialization it will be determined
        according to existing tiles on AWS.

        :return: Index of tile on AWS
        """
        if self.aws_index is not None:
            return self.aws_index
        tile_info_list = get_tile_info(self.tile_name, self.timestamp, all_tiles=True)
        if not tile_info_list:
            raise ValueError("Cannot find aws_index for specified tile and time")

        if self.data_collection is DataCollection.SENTINEL2_L2A:
            for tile_info in sorted(tile_info_list, key=self._parse_aws_index):  # type: ignore
                try:
                    self.aws_index = self._parse_aws_index(tile_info)  # type: ignore
                    self.get_tile_info()
                    return self.aws_index
                except AwsDownloadFailedException:
                    pass

        return self._parse_aws_index(tile_info_list[0])  # type: ignore

    @staticmethod
    def _parse_aws_index(tile_info: dict) -> int:
        """Parses an AWS index from tile info

        :param tile_info: dictionary with information about tile
        :return: Index of tile on AWS
        """
        return int(tile_info["properties"]["s3Path"].split("/")[-1])

    def tile_is_valid(self) -> bool:
        """Checks if tile has tile info and valid timestamp

        :return: `True` if tile is valid and `False` otherwise
        """
        tile_info_datetime = parse_time(self.tile_info["timestamp"], ignoretz=True) if self.tile_info else None
        only_date_given = not isinstance(self.timestamp, dt.datetime)
        return self.tile_info is not None and (only_date_given or self.timestamp == tile_info_datetime)

    def get_tile_info(self) -> dict:
        """
        Collects basic info about tile from tileInfo.json.

        :return: dictionary with tile information
        """
        client = AwsDownloadClient(config=self.config)
        return client.get_json_dict(self.get_url(AwsConstants.TILE_INFO))

    def get_url(self, filename: str) -> str:
        """
        Creates url of file location on AWS.

        :param filename: name of file
        :return: url of file location
        """
        tile_url = self.tile_url
        force_http = filename in [AwsConstants.TILE_INFO, AwsConstants.PRODUCT_INFO, AwsConstants.METADATA]
        if tile_url is None or force_http:
            tile_url = self.get_tile_url(force_http=force_http)
        return f"{tile_url}/{self.add_file_extension(filename)}"

    def get_tile_url(self, force_http: bool = False) -> str:
        """
        Creates base url of tile location on AWS.

        :param force_http: `True` if HTTP base URL should be used and `False` otherwise
        :return: url of tile location
        """
        base_url = self.base_http_url if force_http else self.base_url
        name_parts = self.tile_name[0:2].lstrip("0"), self.tile_name[2], self.tile_name[3:5]

        return (
            f"{base_url}/tiles/{name_parts[0]}/{name_parts[1]}/{name_parts[2]}/{self.date.year}/"
            f"{self.date.month}/{self.date.day}/{self.aws_index}"
        )

    def get_qi_url(self, metafile: str) -> str:
        """Returns url of tile metadata products

        :param metafile: Name of metadata product at AWS
        :return: url location of metadata product at AWS
        """
        return f"{self.tile_url}/qi/{metafile}"

    def get_band_qi_url(self, qi_type: str, band: str = "B00", data_format: MimeType = MimeType.GML) -> str:
        """
        :param qi_type: type of quality indicator
        :param band: band name
        :return: location of gml file on AWS
        """
        band = band.split("/")[-1]
        if data_format == MimeType.JP2:
            return self.get_qi_url(f"{qi_type}_{band}.{data_format.value}")
        return self.get_qi_url(f"MSK_{qi_type}_{band}.{data_format.value}")

    def get_preview_url(self, data_type: str = "L1C") -> str:
        """Returns url location of full resolution L1C preview"""
        if self.data_collection is DataCollection.SENTINEL2_L1C or self.safe_type is EsaSafeType.OLD_TYPE:
            return self.get_url(AwsConstants.PREVIEW_JP2)
        return self.get_qi_url(f"{data_type}_PVI.jp2")

    def get_filepath(self, filename: str) -> str:
        """
        Creates file path for the file.

        :param filename: name of the file
        :return: filename with path on disk
        """
        tile_folder = f"{self.tile_name},{self.date.isoformat()},{self.aws_index}"
        return os.path.join(self.parent_folder, tile_folder, self.add_file_extension(filename)).replace(":", ".")

    def get_product_id(self) -> str:
        """
        Obtains ESA ID of product which contains the tile.

        :return: ESA ID of the product
        """
        return self.tile_info["productName"]

    def _band_exists(self, band_name: str) -> bool:
        if self.data_collection is DataCollection.SENTINEL2_L1C:
            return True
        resolution, band = band_name.split("/")

        if band == "B01" and resolution == AwsConstants.R20m and self.baseline < "04.00":
            return False

        if self.safe_type is EsaSafeType.COMPACT_TYPE:
            return not (band == AwsConstants.VIS and (self.baseline >= "02.07" or self.baseline == "00.01"))

        return band != AwsConstants.TCI and not (band == AwsConstants.SCL and resolution == AwsConstants.R60m)

    @staticmethod
    def tile_id_to_tile(tile_id: str) -> Tuple[str, str, int]:
        """
        :param tile_id: original tile identification string provided by ESA (e.g.
                        'S2A_OPER_MSI_L1C_TL_SGS__20160109T230542_A002870_T10UEV_N02.01')
        :return: tile name, sensing date and AWS index
        """
        if tile_id.split("_")[0] not in ["S2A", "S2B", "L1C"]:
            raise ValueError("Transformation from tile ID to tile works currently only for Sentinel-2 L1C products")

        tile_info = get_tile_info_id(tile_id)
        return AwsData.url_to_tile(tile_info["properties"]["s3Path"])



"""
Module with useful time/date functions
"""
import datetime as dt
from typing import Any, Iterable, List, Optional, Tuple, TypeVar, Union, overload

import dateutil.parser
import dateutil.tz

from .type_utils import RawTimeIntervalType, RawTimeType

TimeType = TypeVar("TimeType", dt.date, dt.datetime)  # pylint: disable=invalid-name


def is_valid_time(time: str) -> bool:
    """Check if input string represents a valid time/date stamp

    :param time: A string containing a time/date.
    :return: `True` is string is a valid time/date, `False` otherwise.
    """
    try:
        dateutil.parser.parse(time)
        return True
    except dateutil.parser.ParserError:
        return False


def parse_time(
    time_input: RawTimeType, *, force_datetime: bool = False, allow_undefined: bool = False, **kwargs: Any
) -> Optional[dt.date]:
    """Parse input time/date string

    :param time_input: An input representation of a time.
    :param force_datetime: If True it will always return datetime.datetime object, if False it can also return only
        `datetime.date` object if only date is provided as input.
    :param allow_undefined: Flag to allow parsing None or '..' into None.
    :param kwargs: Keyword arguments to be passed to `dateutil.parser.parse`. Example: `ignoretz=True`.
    :return: A parsed datetime representing the time.
    """

    if time_input is None or time_input == "..":
        if allow_undefined:
            return None
        raise ValueError("Input is undefined but `allow_undefined` is set to `False`.")

    if isinstance(time_input, dt.date):
        if force_datetime and not isinstance(time_input, dt.datetime):
            return date_to_datetime(time_input)

        if kwargs.get("ignoretz") and isinstance(time_input, dt.datetime):
            return time_input.replace(tzinfo=None)

        return time_input

    time = dateutil.parser.parse(time_input, **kwargs)
    if force_datetime or len(time_input) > 10:  # This check is not very accurate, but it works for ISO format
        return time
    return time.date()


def parse_time_interval(
    time: Union[RawTimeType, RawTimeIntervalType], allow_undefined: bool = False, **kwargs: Any
) -> Tuple[Optional[dt.datetime], Optional[dt.datetime]]:
    """Parse input into an interval of two times, specifying start and end time, into datetime objects.

    The input time can have the following formats, which will be parsed as:

    * `YYYY-MM-DD` -> `[YYYY-MM-DD:T00:00:00, YYYY-MM-DD:T23:59:59]`
    * `YYYY-MM-DDThh:mm:ss` -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`
    * list or tuple of two dates in form `YYYY-MM-DD` -> `[YYYY-MM-DDT00:00:00, YYYY-MM-DDT23:59:59]`
    * list or tuple of two dates in form `YYYY-MM-DDThh:mm:ss` -> `[YYYY-MM-DDThh:mm:ss, YYYY-MM-DDThh:mm:ss]`

    All input times can also be specified as `datetime` objects. Instances of `datetime.date` will be treated as
    `YYYY-MM-DD` and instance of `datetime.datetime` will be treated as `YYYY-MM-DDThh:mm:ss`.

    :param time: An input representation of a time interval.
    :param allow_undefined: Boolean flag controls if None or '..' are allowed.
    :param kwargs: Keyword arguments to be passed to `parse_time` function.
    :return: A pair of datetime objects defining the time interval.
    :raises: ValueError
    """
    date_interval: Tuple[Union[None, dt.date], Union[None, dt.date]]

    if allow_undefined and time in [None, ".."]:
        date_interval = None, None
    elif isinstance(time, (str, dt.date)):
        parsed_time = parse_time(time, **kwargs)
        date_interval = parsed_time, parsed_time
    elif isinstance(time, (tuple, list)) and len(time) == 2:
        date_interval = parse_time(time[0], allow_undefined=allow_undefined, **kwargs), parse_time(
            time[1], allow_undefined=allow_undefined, **kwargs
        )
    else:
        raise ValueError("Time must be a string/datetime object or tuple/list of 2 strings/datetime objects")

    start_time, end_time = date_interval
    if not isinstance(start_time, dt.datetime) and start_time is not None:
        start_time = date_to_datetime(start_time)
    if not isinstance(end_time, dt.datetime) and end_time is not None:
        end_time = date_to_datetime(end_time, dt.time(hour=23, minute=59, second=59))

    if start_time and end_time and start_time > end_time:
        raise ValueError("Start of time interval is larger than end of time interval")

    return start_time, end_time


@overload
def serialize_time(timestamp_input: Optional[dt.date], *, use_tz: bool = False) -> str:
    ...


@overload
def serialize_time(timestamp_input: Iterable[Optional[dt.date]], *, use_tz: bool = False) -> Tuple[str, ...]:
    ...


def serialize_time(
    timestamp_input: Union[None, dt.date, Iterable[Optional[dt.date]]], *, use_tz: bool = False
) -> Union[str, Tuple[str, ...]]:
    """Transforms datetime objects into ISO 8601 strings.

    :param timestamp_input: A datetime object or a tuple of datetime objects.
    :param use_tz: If `True` it will ensure that the serialized string contains a timezone information (typically
        with `Z` at the end instead of +00:00). If `False` it will make sure to remove any timezone information.
    :return: Timestamp(s) serialized into string(s).
    """
    if isinstance(timestamp_input, Iterable):
        return tuple(serialize_time(timestamp, use_tz=use_tz) for timestamp in timestamp_input)

    if timestamp_input is None:
        return ".."

    if not isinstance(timestamp_input, dt.date):
        raise ValueError("Expected a datetime object or a tuple of datetime objects")

    if use_tz:
        if not isinstance(timestamp_input, dt.datetime):
            raise ValueError(
                "Cannot ensure timezone information for datetime.date objects, use datetime.datetime instead"
            )

        if not timestamp_input.tzinfo:
            timestamp_input = timestamp_input.replace(tzinfo=dateutil.tz.tzutc())

    elif isinstance(timestamp_input, dt.datetime) and timestamp_input.tzinfo:
        timestamp_input = timestamp_input.replace(tzinfo=None)

    return timestamp_input.isoformat().replace("+00:00", "Z")


def date_to_datetime(date: dt.date, time: Optional[dt.time] = None) -> dt.datetime:
    """Converts a date object into datetime object.

    :param date: A date object.
    :param time: An option time object, if not provided it will replace it with `00:00:00`.
    :return: A datetime object derived from date and time.
    """
    if time is None:
        time = dt.datetime.min.time()
    return dt.datetime.combine(date, time)


def filter_times(timestamps: Iterable[TimeType], time_difference: dt.timedelta) -> List[TimeType]:
    """Filters out timestamps within time_difference, preserving only the oldest timestamp.

    :param timestamps: A list of timestamps.
    :param time_difference: A time difference threshold.
    :return: An ordered list of timestamps `d_1 <= d_2 <= ... <= d_n` such that `d_(i+1)-d_i > time_difference`.
    """
    timestamps = sorted(set(timestamps))

    filtered_timestamps: List[TimeType] = []
    for current_timestamp in timestamps:
        if not filtered_timestamps or current_timestamp - filtered_timestamps[-1] > time_difference:
            filtered_timestamps.append(current_timestamp)

    return filtered_timestamps



"""
Module implementing an interface with
`Sentinel Hub Bring Your Own COG API <https://docs.sentinel-hub.com/api/latest/api/byoc/>`__.
"""
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, Optional, Union

from dataclasses_json import CatchAll, LetterCase, Undefined
from dataclasses_json import config as dataclass_config
from dataclasses_json import dataclass_json

from ..constants import MimeType, RequestType
from ..data_collections import DataCollection
from ..geometry import Geometry
from ..type_utils import Json, JsonDict
from .base import BaseCollection, SentinelHubFeatureIterator, SentinelHubService
from .utils import datetime_config, geometry_config, remove_undefined

CollectionType = Union["ByocCollection", DataCollection, dict, str]
TileType = Union["ByocTile", dict, str]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class ByocCollectionBand:
    """Dataclass to hold BYOC collection band specification"""

    other_data: CatchAll
    source: Optional[str] = None
    band_index: Optional[int] = None
    bit_depth: int = 8
    sample_format: str = "UINT"
    no_data: Optional[float] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class ByocCollectionAdditionalData:
    """Dataclass to hold BYOC collection additional data"""

    other_data: CatchAll
    bands: Optional[Dict[str, ByocCollectionBand]] = None
    max_meters_per_pixel: Optional[float] = None
    max_meters_per_pixel_override: Optional[float] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class ByocCollection(BaseCollection):
    """Dataclass to hold BYOC collection data"""

    additional_data: Optional[ByocCollectionAdditionalData] = None


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class ByocTile:
    """Dataclass to hold BYOC tile data"""

    path: str
    other_data: CatchAll
    status: Optional[str] = None
    tile_id: Optional[str] = field(metadata=dataclass_config(field_name="id"), default=None)
    tile_geometry: Optional[Geometry] = field(metadata=geometry_config, default=None)
    cover_geometry: Optional[Geometry] = field(metadata=geometry_config, default=None)
    created: Optional[datetime] = field(metadata=datetime_config, default=None)
    sensing_time: Optional[datetime] = field(metadata=datetime_config, default=None)
    additional_data: Optional[dict] = None


class SentinelHubBYOC(SentinelHubService):
    """An interface class for Sentinel Hub Bring your own COG (BYOC) API

    For more info check `BYOC API reference
    <https://docs.sentinel-hub.com/api/latest/reference/#tag/byoc_collection>`__.
    """

    @staticmethod
    def _get_service_url(base_url: str) -> str:
        """Provides URL to Catalog API"""
        return f"{base_url}/api/v1/byoc"

    def iter_collections(self, search: Optional[str] = None, **kwargs: Any) -> SentinelHubFeatureIterator:
        """Retrieve collections

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollections>`__

        :param search: A search query
        :param kwargs: Any other request parameters
        :return: iterator over collections
        """
        return SentinelHubFeatureIterator(
            client=self.client,
            url=f"{self.service_url}/collections",
            params={"search": search, **kwargs},
            exception_message="Failed to obtain information about available BYOC collections",
        )

    def get_collection(self, collection: CollectionType) -> JsonDict:
        """Get collection by its id

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollectionById>`__

        :param collection: a ByocCollection, dict or collection id string
        :return: dictionary of the collection
        """
        url = f"{self.service_url}/collections/{self._parse_id(collection)}"
        return self.client.get_json_dict(url=url, use_session=True, extract_key="data")

    def create_collection(self, collection: CollectionType) -> JsonDict:
        """Create a new collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/createByocCollection>`__

        :param collection: ByocCollection object or a dictionary
        :return: dictionary of the created collection
        """
        coll = self._to_dict(collection)
        url = f"{self.service_url}/collections"
        return self.client.get_json_dict(url=url, post_values=coll, use_session=True, extract_key="data")

    def update_collection(self, collection: CollectionType) -> Json:
        """Update an existing collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/updateByocCollectionById>`__

        :param collection: ByocCollection object or a dictionary
        """
        coll = self._to_dict(collection)
        url = f"{self.service_url}/collections/{self._parse_id(coll)}"
        headers = {"Content-Type": MimeType.JSON.get_string()}
        return self.client.get_json(
            url=url, request_type=RequestType.PUT, post_values=coll, headers=headers, use_session=True
        )

    def delete_collection(self, collection: CollectionType) -> Json:
        """Delete existing collection by its id

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/deleteByocCollectionById>`__

        :param collection: a ByocCollection, dict or collection id string
        """
        url = f"{self.service_url}/collections/{self._parse_id(collection)}"
        return self.client.get_json(url=url, request_type=RequestType.DELETE, use_session=True)

    def copy_tiles(self, from_collection: CollectionType, to_collection: CollectionType) -> Json:
        """Copy tiles from one collection to another

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/copyByocCollectionTiles>`__

        :param from_collection: a ByocCollection, dict or collection id string
        :param to_collection: a ByocCollection, dict or collection id string
        """
        url = (
            f"{self.service_url}/collections/{self._parse_id(from_collection)}"
            f"/copyTiles?toCollection={self._parse_id(to_collection)}"
        )
        return self.client.get_json(url=url, request_type=RequestType.POST, use_session=True)

    def iter_tiles(
        self, collection: CollectionType, sort: Optional[str] = None, path: Optional[str] = None, **kwargs: Any
    ) -> SentinelHubFeatureIterator:
        """Iterator over collection tiles

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollectionTiles>`__

        :param collection: a ByocCollection, dict or collection id string
        :param sort: Order in which to return tiles
        :param path: An exact path where tiles are located
        :param kwargs: Any other request parameters
        :return: An iterator over payloads of tiles from the collection
        """
        collection_id = self._parse_id(collection)
        return SentinelHubFeatureIterator(
            client=self.client,
            url=f"{self.service_url}/collections/{collection_id}/tiles",
            params={"sort": sort, "path": path, **kwargs},
            exception_message=f"Failed to obtain information about tiles in BYOC collection {collection_id}",
        )

    def get_tile(self, collection: CollectionType, tile: TileType) -> JsonDict:
        """Get a tile of collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/getByocCollectionTileById>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile, dict or tile id string
        :return: dictionary of the tile
        """
        url = f"{self.service_url}/collections/{self._parse_id(collection)}/tiles/{self._parse_id(tile)}"
        return self.client.get_json_dict(url=url, use_session=True, extract_key="data")

    def create_tile(self, collection: CollectionType, tile: TileType) -> JsonDict:
        """Create tile within collection

        `BYOC API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/createByocCollectionTile>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile or dict
        :return: dictionary of the tile
        """
        _tile = self._to_dict(tile)
        url = f"{self.service_url}/collections/{self._parse_id(collection)}/tiles"
        return self.client.get_json_dict(url=url, post_values=_tile, use_session=True, extract_key="data")

    def update_tile(self, collection: CollectionType, tile: TileType) -> Json:
        """Update a tile within collection

        `BYOC API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/updateByocCollectionTileById>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile or dict
        """
        url = f"{self.service_url}/collections/{self._parse_id(collection)}/tiles/{self._parse_id(tile)}"
        headers = {"Content-Type": MimeType.JSON.get_string()}

        _tile = self._to_dict(tile)
        updates = remove_undefined(
            {
                "path": _tile["path"],
                "coverGeometry": _tile.get("coverGeometry"),
                "sensingTime": _tile.get("sensingTime"),
            }
        )

        return self.client.get_json(
            url=url, request_type=RequestType.PUT, post_values=updates, headers=headers, use_session=True
        )

    def delete_tile(self, collection: CollectionType, tile: TileType) -> Json:
        """Delete a tile from collection

        `BYOC API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/deleteByocCollectionTileById>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile, dict or tile id string
        """
        url = f"{self.service_url}/collections/{self._parse_id(collection)}/tiles/{self._parse_id(tile)}"
        return self.client.get_json(url=url, request_type=RequestType.DELETE, use_session=True)

    def reingest_tile(self, collection: CollectionType, tile: TileType) -> Json:
        """Re-ingests a tile into a collection

        `BYOC API reference
        <https://docs.sentinel-hub.com/api/latest/reference/#operation/reingestByocCollectionTileById>`__

        :param collection: a ByocCollection, dict or collection id string
        :param tile: a ByocTile, dict or tile id string
        """
        url = f"{self.service_url}/collections/{self._parse_id(collection)}/tiles/{self._parse_id(tile)}/reingest"
        return self.client.get_json(url=url, request_type=RequestType.POST, use_session=True)

    @staticmethod
    def _parse_id(data: object) -> Optional[str]:
        if isinstance(data, (ByocCollection, DataCollection)):
            return data.collection_id
        if isinstance(data, ByocTile):
            return data.tile_id
        if isinstance(data, dict):
            return data["id"]
        if isinstance(data, str):
            return data
        raise ValueError(f"Expected a BYOC/Data dataclass, dictionary or a string, got {data}.")

    @staticmethod
    def _to_dict(data: object) -> dict:
        """Constructs dict from an object (either dataclass or dict)"""
        if isinstance(data, (ByocCollection, ByocTile, ByocCollectionAdditionalData, ByocCollectionBand)):
            return data.to_dict()  # type: ignore[union-attr]
        if isinstance(data, dict):
            return data
        raise ValueError(f"Expected either a data class (e.g., ByocCollection and similar) or a dict, got {data}.")



"""
Module with custom types and utilities.
"""
import datetime as dt
from typing import Any, Dict, Tuple, Union

RawTimeType = Union[None, str, dt.date]
RawTimeIntervalType = Tuple[RawTimeType, RawTimeType]
JsonDict = Dict[str, Any]
Json = Union[JsonDict, list, str, float, int, None]



"""
Module implementing some utility functions not suitable for other utility modules
"""
import sys
import warnings
from abc import ABCMeta
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, Iterable, Optional, Union
from urllib.parse import urlencode

from dataclasses_json import CatchAll, LetterCase, Undefined
from dataclasses_json import config as dataclass_config
from dataclasses_json import dataclass_json

from ..base import FeatureIterator
from ..config import SHConfig
from ..data_collections import DataCollection
from ..download.sentinelhub_client import SentinelHubDownloadClient
from ..exceptions import MissingDataInRequestException, SHDeprecationWarning
from ..type_utils import JsonDict
from .utils import datetime_config, remove_undefined

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol  # pylint: disable=ungrouped-imports


class SentinelHubService(metaclass=ABCMeta):
    """A base class for classes interacting with different Sentinel Hub APIs"""

    def __init__(self, config: Optional[SHConfig] = None, base_url: Optional[str] = None):
        """
        :param config: A configuration object with required parameters `sh_client_id`, `sh_client_secret`, and
            `sh_auth_base_url` which is used for authentication and `sh_base_url` which defines the service
            deployment that will be used.
        :param base_url: A deprecated parameter. Use `config` instead.
        """
        self.config = config or SHConfig()

        if base_url:
            warnings.warn(
                "Parameter base_url is deprecated and will soon be removed. Instead set "
                "config.sh_base_url = base_url and provide it with config parameter",
                category=SHDeprecationWarning,
            )

        base_url = base_url or self.config.sh_base_url
        if not isinstance(base_url, str):
            raise ValueError(f"Sentinel Hub base URL parameter should be a string but got {base_url}")
        base_url = base_url.rstrip("/")
        self.service_url = self._get_service_url(base_url)

        self.client = SentinelHubDownloadClient(config=self.config)

    @staticmethod
    def _get_service_url(base_url: str) -> str:
        """Provides the URL to a specific service"""


class SentinelHubFeatureIterator(FeatureIterator[JsonDict]):
    """Feature iterator for the most common implementation of feature pagination at Sentinel Hub services"""

    def __init__(self, *args: Any, exception_message: Optional[str] = None, **kwargs: Any):
        """
        :param args: Arguments passed to FeatureIterator
        :param exception_message: A message to be raised if no features are found
        :param kwargs: Keyword arguments passed to FeatureIterator
        """
        self.exception_message = exception_message or "No data found"
        self.next: Optional[JsonDict] = None

        super().__init__(*args, **kwargs)

    def _fetch_features(self) -> Iterable[JsonDict]:
        """Collect more results from the service"""
        params = remove_undefined({**self.params, "viewtoken": self.next})
        url = f"{self.url}?{urlencode(params)}"

        json_response = self.client.get_json_dict(url, use_session=True)

        new_features = json_response.get("data")
        if new_features is None:
            raise MissingDataInRequestException(self.exception_message)

        self.next = json_response["links"].get("nextToken")
        self.finished = self.next is None or not new_features

        return new_features


class _AdditionalData(Protocol):
    """Describes minimum requirements for additional data passed to BaseCollection"""

    bands: Optional[Dict[str, Any]]


@dataclass_json(letter_case=LetterCase.CAMEL, undefined=Undefined.INCLUDE)
@dataclass
class BaseCollection:
    """Dataclass to hold data about a collection"""

    name: str
    s3_bucket: str
    other_data: CatchAll
    additional_data: Optional[_AdditionalData]
    collection_id: Optional[str] = field(metadata=dataclass_config(field_name="id"), default=None)
    user_id: Optional[str] = None
    created: Optional[datetime] = field(metadata=datetime_config, default=None)
    no_data: Optional[Union[int, float]] = None

    def to_data_collection(self) -> DataCollection:
        """Returns a DataCollection enum for this collection"""
        if self.collection_id is None:
            raise ValueError("This collection is missing a collection id")

        if self.additional_data and self.additional_data.bands:
            band_names = tuple(self.additional_data.bands)
        else:
            band_names = None

        return DataCollection.define_byoc(collection_id=self.collection_id, bands=band_names)



"""
Module for creating .SAFE structure with data collected from AWS
"""

import warnings
from typing import Any, List, Optional, Tuple

from ..constants import MimeType
from ..data_collections import DataCollection
from ..download.request import DownloadRequest
from ..exceptions import SHRuntimeWarning
from .client import AwsDownloadClient
from .constants import AwsConstants, EsaSafeType
from .data import AwsProduct, AwsTile


class SafeProduct(AwsProduct):
    """Class implementing transformation of Sentinel-2 satellite products from AWS into .SAFE structure"""

    def get_requests(self) -> Tuple[List[DownloadRequest], List[str]]:
        """Creates product structure and returns list of files for download

        :return: list of download requests
        """
        safe = self.get_safe_struct()

        self.download_list = []
        self.structure_recursion(safe, self.parent_folder)
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_safe_struct(self) -> dict:
        """Describes a structure inside tile folder of ESA product .SAFE structure

        :return: nested dictionaries representing .SAFE structure
        """
        main_folder = self.get_main_folder()
        safe = {
            main_folder: {
                AwsConstants.AUX_DATA: {},
                AwsConstants.DATASTRIP: self._get_datastrip_substruct(),
                AwsConstants.GRANULE: self._get_granule_substruct(),
                self.get_product_metadata_name(): self.get_url(AwsConstants.METADATA),
                "INSPIRE.xml": self.get_url(AwsConstants.INSPIRE),
                self.add_file_extension(AwsConstants.MANIFEST): self.get_url(AwsConstants.MANIFEST),
                AwsConstants.HTML: {},  # AWS doesn't have this data
                AwsConstants.INFO: {},  # AWS doesn't have this data
            }
        }

        if self.is_early_compact_l2a():
            safe[main_folder]["L2A_Manifest.xml"] = self.get_url(AwsConstants.L2A_MANIFEST)
            safe[main_folder][self.get_report_name()] = self.get_url(AwsConstants.REPORT)

        if self.safe_type is EsaSafeType.OLD_TYPE and self.baseline != "02.02":
            safe[main_folder][_edit_name(self.product_id, "BWI") + ".png"] = self.get_url(
                AwsConstants.PREVIEW, MimeType.PNG
            )
        return safe

    def _get_datastrip_substruct(self) -> dict:
        """Builds a datastrip subfolder structure of .SAFE format."""
        datastrip_safe: dict = {}
        datastrip_list = self.get_datastrip_list()
        for datastrip_folder, datastrip_url in datastrip_list:
            datastrip_safe[datastrip_folder] = {AwsConstants.QI_DATA: {}}

            if self.has_reports():
                for metafile in AwsConstants.QUALITY_REPORTS:
                    metafile_name = self.add_file_extension(metafile)

                    metafile_s3_name = metafile_name
                    # S-2 L1C reports are available on S3 under modified names
                    if self.data_collection is DataCollection.SENTINEL2_L1C:
                        metafile_s3_name = metafile_s3_name.replace(".xml", "_report.xml")

                    metafile_url = f"{datastrip_url}/qi/{metafile_s3_name}"
                    datastrip_safe[datastrip_folder][AwsConstants.QI_DATA][metafile_name] = metafile_url

            datastrip_name = self.get_datastrip_metadata_name(datastrip_folder)
            datastrip_url = f"{datastrip_url}/{self.add_file_extension(AwsConstants.METADATA)}"
            datastrip_safe[datastrip_folder][datastrip_name] = datastrip_url

        return datastrip_safe

    def _get_granule_substruct(self) -> dict:
        """Builds a granule subfolder structure of .SAFE format."""
        granule_safe = {}
        for tile_info in self.product_info["tiles"]:
            tile_name, date, aws_index = self.url_to_tile(self.get_tile_url(tile_info))

            if self.tile_list is None or AwsTile.parse_tile_name(tile_name) in self.tile_list:
                tile_struct = SafeTile(
                    tile_name,
                    date,
                    aws_index,
                    parent_folder=None,
                    bands=self.bands,
                    metafiles=self.metafiles,
                    data_collection=self.data_collection,
                ).get_safe_struct()

                granule_safe.update(tile_struct)

        return granule_safe

    def get_main_folder(self) -> str:
        """
        :return: name of main folder
        """
        return f"{self.product_id}.SAFE"

    def get_datastrip_list(self) -> List[Tuple[str, str]]:
        """
        :return: list of datastrips folder names and urls from `productInfo.json` file
        """
        datastrips = self.product_info["datastrips"]
        return [
            (self.get_datastrip_name(datastrip["id"]), f'{self.base_url}/{datastrip["path"]}')
            for datastrip in datastrips
        ]

    def get_datastrip_name(self, datastrip: str) -> str:
        """
        :param datastrip: name of datastrip
        :return: name of datastrip folder
        """
        if self.safe_type is EsaSafeType.OLD_TYPE:
            return datastrip
        return "_".join(datastrip.split("_")[4:-1])

    def get_datastrip_metadata_name(self, datastrip_folder: str) -> str:
        """
        :param datastrip_folder: name of datastrip folder
        :return: name of datastrip metadata file
        """
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = _edit_name(datastrip_folder, "MTD", delete_end=True)
        else:
            name = "MTD_DS"
        return f"{name}.{MimeType.XML.value}"

    def get_product_metadata_name(self) -> str:
        """
        :return: name of product metadata file
        """
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = _edit_name(self.product_id, "MTD", "SAFL1C")
        else:
            name = f'MTD_{self.product_id.split("_")[1]}'
        return f"{name}.{MimeType.XML.value}"

    def get_report_name(self) -> str:
        """
        :return: name of the report file of L2A products
        """
        return f"{self.product_id}_{self.get_report_time()}_report.{MimeType.XML.value}"

    def get_report_time(self) -> str:
        """Returns time when the L2A processing started and reports was created.
        :return: String in a form YYYYMMDDTHHMMSS
        """
        client = AwsDownloadClient(config=self.config)
        tree = client.get_xml(self.get_url(AwsConstants.REPORT))

        try:
            timestamp = tree.find("check/inspection").attrib["execution"]  # type: ignore
            return timestamp.split(",")[0].replace(" ", "T").replace(":", "").replace("-", "")
        except AttributeError:
            warnings.warn("Could not obtain the L2A report creation time", category=SHRuntimeWarning)
            return "unknown"


class SafeTile(AwsTile):
    """Class implementing transformation of Sentinel-2 satellite tiles from AWS into .SAFE structure"""

    def __init__(self, *args: Any, **kwargs: Any):
        """Initialization parameters are inherited from parent class"""
        super().__init__(*args, **kwargs)

        self.tile_id = self.get_tile_id()

    def get_requests(self) -> Tuple[List[DownloadRequest], List[str]]:
        """Creates tile structure and returns list of files for download.

        :return: list of download requests for
        """
        safe = self.get_safe_struct()

        self.download_list = []
        self.structure_recursion(safe, self.parent_folder)
        self.sort_download_list()
        return self.download_list, self.folder_list

    def get_safe_struct(self) -> dict:
        """Describes a structure inside tile folder of ESA product .SAFE structure.

        :return: nested dictionaries representing .SAFE structure
        """
        return {
            self.get_main_folder(): {
                AwsConstants.AUX_DATA: self._get_aux_substruct(),
                AwsConstants.IMG_DATA: self._get_image_substruct(),
                AwsConstants.QI_DATA: self._get_qi_substruct(),
                self.get_tile_metadata_name(): self.get_url(AwsConstants.METADATA),
            }
        }

    def _get_aux_substruct(self) -> dict:
        """Builds an auxiliary data subfolder structure of .SAFE format.

        Note: Old products also have DEM and MSI in aux folder which are not reconstructed here.
        """
        aux_safe = {}

        # Not sure if 2nd condition of the following is correct:
        if self.data_collection is DataCollection.SENTINEL2_L2A or self.baseline != "02.04":
            ecmwft_file = (
                AwsConstants.ECMWFT
                if self.data_collection is DataCollection.SENTINEL2_L1C or self.safe_type is EsaSafeType.OLD_TYPE
                else AwsConstants.AUX_ECMWFT
            )
            aux_safe[self.get_aux_data_name()] = self.get_url(ecmwft_file)

            if self.baseline >= "04.00":
                camsfo_file = (
                    AwsConstants.CAMSFO
                    if self.data_collection is DataCollection.SENTINEL2_L1C
                    else AwsConstants.AUX_CAMSFO
                )
                aux_safe["AUX_CAMSFO"] = self.get_url(camsfo_file)

        if self.is_early_compact_l2a():
            gipp_filename = self.add_file_extension(AwsConstants.GIPP, remove_path=True)
            aux_safe[gipp_filename] = self.get_url(AwsConstants.GIPP)

        return aux_safe

    def _get_image_substruct(self) -> dict:
        """Builds the part of structure of .SAFE format that contains satellite imagery."""
        img_safe: dict = {}

        if self.data_collection is DataCollection.SENTINEL2_L1C:
            for band in self.bands:
                img_safe[self.get_img_name(band)] = self.get_url(band)

            if self.safe_type is EsaSafeType.COMPACT_TYPE:
                img_safe[self.get_img_name(AwsConstants.TCI)] = self.get_url(AwsConstants.TCI)

            return img_safe

        for resolution in AwsConstants.RESOLUTIONS:
            img_safe[resolution] = {}

        for band_name in self.bands:
            resolution, band = band_name.split("/")
            if self._band_exists(band_name):
                img_safe[resolution][self.get_img_name(band, resolution)] = self.get_url(band_name)

        return img_safe

    def _get_qi_substruct(self) -> dict:
        """Builds a quality-indicators data subfolder structure of .SAFE format."""
        qi_safe = self._get_reports_substruct()

        if self.baseline >= "04.00":
            qi_list = AwsConstants.QI_LIST_V4
            qi_data_format = MimeType.JP2
            clouds_qi_name = "CLASSI"
        else:
            qi_list = AwsConstants.QI_LIST
            qi_data_format = MimeType.GML
            clouds_qi_name = "CLOUDS"

        clouds_qi_filename = self.get_qi_name(clouds_qi_name, data_format=qi_data_format)
        qi_safe[clouds_qi_filename] = self.get_band_qi_url(clouds_qi_name, data_format=qi_data_format)

        for qi_type in qi_list:
            for band in AwsConstants.S2_L1C_BANDS:
                band_qi_filename = self.get_qi_name(qi_type, band, data_format=qi_data_format)
                qi_safe[band_qi_filename] = self.get_band_qi_url(qi_type, band, data_format=qi_data_format)

        if self.data_collection is DataCollection.SENTINEL2_L2A:
            for mask in AwsConstants.CLASS_MASKS:
                for resolution in [AwsConstants.R20m, AwsConstants.R60m]:
                    if "00.01" < self.baseline <= "02.06":
                        mask_name = self.get_img_name(mask, resolution)
                    else:
                        mask_name = self.get_qi_name(f"{mask}PRB", resolution.lstrip("R"), MimeType.JP2)
                    qi_safe[mask_name] = self.get_qi_url(f'{mask}_{resolution.lstrip("R")}.jp2')

        if self.is_early_compact_l2a():
            qi_safe[self.get_img_name(AwsConstants.PVI)] = self.get_preview_url("L2A")

        is_newer_l2a_version = self.data_collection is DataCollection.SENTINEL2_L2A and (
            self.baseline >= "02.07" or self.baseline == "00.01"
        )
        preview_type = "L2A" if is_newer_l2a_version else "L1C"
        qi_safe[self.get_preview_name()] = self.get_preview_url(preview_type)

        return qi_safe

    def _get_reports_substruct(self) -> dict:
        """Builds a substructure of .SAFE format with reports."""
        reports_safe: dict = {}
        if not self.has_reports():
            return reports_safe

        reports = AwsConstants.QUALITY_REPORTS
        if self.data_collection is DataCollection.SENTINEL2_L2A and self.baseline >= "04.00":
            reports = reports + [AwsConstants.L2A_QUALITY]

        for metafile in reports:
            if (
                metafile == AwsConstants.RADIOMETRIC_QUALITY
                and self.data_collection is DataCollection.SENTINEL2_L2A
                and (self.baseline <= "02.07" or self.baseline >= "04.00")
            ):
                continue

            metafile_name = self.add_file_extension(metafile, remove_path=True)
            reports_safe[metafile_name] = self.get_qi_url(metafile_name)

        return reports_safe

    def get_tile_id(self) -> str:
        """Creates ESA tile ID

        :return: ESA tile ID
        """
        client = AwsDownloadClient(config=self.config)
        tree = client.get_xml(self.get_url(AwsConstants.METADATA))

        tile_id_tag = (
            "TILE_ID_2A"
            if (self.data_collection is DataCollection.SENTINEL2_L2A and "00.01" < self.baseline <= "02.06")
            else "TILE_ID"
        )
        tile_id = tree[0].find(tile_id_tag).text  # type: ignore
        if self.safe_type is EsaSafeType.OLD_TYPE:
            return tile_id

        info = tile_id.split("_")

        if (self.data_collection is DataCollection.SENTINEL2_L1C and self.baseline >= "02.07") or (
            self.data_collection is DataCollection.SENTINEL2_L2A
            and (self.baseline >= "02.10" or self.baseline == "00.01")
        ):
            tile_id_time = self.get_datastrip_time()
        else:
            tile_id_time = self.get_sensing_time()

        return "_".join([info[3], info[-2], info[-3], tile_id_time])

    def get_sensing_time(self) -> str:
        """
        :return: Exact tile sensing time
        """
        return self.tile_info["timestamp"].split(".")[0].replace("-", "").replace(":", "")

    def get_datastrip_time(self) -> str:
        """
        :return: Exact datastrip time
        """
        # S2A_OPER_MSI_L1C_DS_EPAE_20181119T061056_S20181119T031012_N02.07 -> 20181119T031012
        # S2A_OPER_MSI_L1C_DS_EPA__20190225T132350_S20190129T144524_N02.07 -> 20190129T144524
        return self.tile_info["datastrip"]["id"].replace("__", "_").split("_")[7].lstrip("S")

    def get_datatake_time(self) -> str:
        """
        :return: Exact time of datatake
        """
        return self.tile_info["productName"].split("_")[2]

    def get_main_folder(self) -> str:
        """
        :return: name of tile folder
        """
        return self.tile_id

    def get_tile_metadata_name(self) -> str:
        """
        :return: name of tile metadata file
        """
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = _edit_name(self.tile_id, "MTD", delete_end=True)
        else:
            name = "MTD_TL"
        return f"{name}.xml"

    def get_aux_data_name(self) -> str:
        """
        :return: name of auxiliary data file
        """
        if self.safe_type is EsaSafeType.OLD_TYPE:
            # this is not correct, but we cannot reconstruct last two timestamps in auxiliary data file name
            # e.g. S2A_OPER_AUX_ECMWFT_EPA__20160120T231011_V20160103T150000_20160104T030000
            return "AUX_ECMWFT"
        return "AUX_ECMWFT"

    def get_img_name(self, band: str, resolution: Optional[str] = None) -> str:
        """
        :param band: band name
        :param resolution: Specifies the resolution in case of Sentinel-2 L2A products
        :return: name of band image file
        """
        band = band.split("/")[-1]
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = self.tile_id.rsplit("_", 1)[0] + "_" + band
        else:
            name = "_".join([self.tile_id.split("_")[1], self.get_datatake_time(), band])
        if self.data_collection is DataCollection.SENTINEL2_L2A and resolution is not None:
            name = f'{name}_{resolution.lstrip("R")}'
        if self.data_collection is DataCollection.SENTINEL2_L2A and "00.01" < self.baseline <= "02.06":
            name = f"L2A_{name}"
        return f"{name}.jp2"

    def get_qi_name(self, qi_type: str, band: str = "B00", data_format: MimeType = MimeType.GML) -> str:
        """
        :param qi_type: type of quality indicator
        :param band: band name
        :param data_format: format of the file
        :return: name of gml file
        """
        band = band.split("/")[-1]
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = _edit_name(self.tile_id, "MSK", delete_end=True)
            collection_param = f"{'L1C' if self.data_collection is DataCollection.SENTINEL2_L1C else 'L2A'}_TL"
            name = name.replace(collection_param, qi_type)
            name = f"{name}_{band}_MSIL1C"
        else:
            name = f"MSK_{qi_type}_{band}"
        return f"{name}.{data_format.value}"

    def get_preview_name(self) -> str:
        """Returns .SAFE name of full resolution L1C preview
        :return: name of preview file
        """
        if self.safe_type is EsaSafeType.OLD_TYPE:
            name = _edit_name(self.tile_id, AwsConstants.PVI, delete_end=True)
        else:
            name = "_".join([self.tile_id.split("_")[1], self.get_datatake_time(), AwsConstants.PVI])
        return f"{name}.jp2"


def _edit_name(name: str, code: str, add_code: Optional[str] = None, delete_end: bool = False) -> str:
    """Helping function for creating file names in .SAFE format

    :param name: initial string
    :param code:
    :param add_code:
    :param delete_end:
    :return: edited string
    """
    info = name.split("_")
    info[2] = code
    if add_code is not None:
        info[3] = add_code
    if delete_end:
        info.pop()
    return "_".join(info)



"""
Module for data decoding
"""
import json
import struct
import tarfile
import warnings
from io import BytesIO, IOBase
from json import JSONDecodeError
from typing import Any, Dict, Union
from xml.etree import ElementTree

import numpy as np
import tifffile as tiff
from PIL import Image
from requests import Response

from .constants import MimeType
from .exceptions import ImageDecodingError


def decode_data(response_content: bytes, data_type: MimeType) -> Any:
    """Interprets downloaded data and returns it.

    :param response_content: downloaded data (i.e. json, png, tiff, xml, zip, ... file)
    :param data_type: expected downloaded data type
    :return: downloaded data
    :raises: ValueError
    """
    if data_type is MimeType.JSON:
        response_text = response_content.decode("utf-8")
        if not response_text:
            return response_text
        return json.loads(response_text)
    if data_type is MimeType.TAR:
        return decode_tar(response_content)
    if MimeType.is_image_format(data_type):
        return decode_image(response_content, data_type)
    if data_type is MimeType.XML or data_type is MimeType.GML or data_type is MimeType.SAFE:
        return ElementTree.fromstring(response_content)

    try:
        return {
            MimeType.RAW: response_content,
            MimeType.TXT: response_content,
            MimeType.ZIP: BytesIO(response_content),
        }[data_type]
    except KeyError as exception:
        raise ValueError(f"Decoding data format {data_type} is not supported") from exception


def decode_image(data: bytes, image_type: MimeType) -> np.ndarray:
    """Decodes the image provided in various formats, i.e. png, 16-bit float tiff, 32-bit float tiff, jp2
    and returns it as a numpy array

    :param data: image in its original format
    :param image_type: expected image format
    :return: image as numpy array
    :raises: ImageDecodingError
    """
    bytes_data = BytesIO(data)
    if image_type is MimeType.TIFF:
        image = tiff.imread(bytes_data)
    elif image_type is MimeType.JP2:
        image = decode_jp2_image(bytes_data)
    else:
        image = decode_image_with_pillow(bytes_data)

    if image is None:
        raise ImageDecodingError("Unable to decode image")
    return image


def decode_image_with_pillow(stream: Union[IOBase, str]) -> np.ndarray:
    """Decodes an image using `Pillow` package and handles potential warnings.

    :param stream: A binary stream format or a filename.
    :return: A numpy array representing an image of shape (height, width) or (height, width, channels).
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", Image.DecompressionBombWarning)
        return np.array(Image.open(stream))


def decode_jp2_image(stream: IOBase) -> np.ndarray:
    """Tries to decode a JPEG2000 image using the `Pillow` package.

    :param stream: A binary stream format.
    :return: A numpy array representing an image of shape (height, width) or (height, width, channels).
    """
    image = decode_image_with_pillow(stream)
    bit_depth = get_jp2_bit_depth(stream)

    return fix_jp2_image(image, bit_depth)


def decode_tar(data: Union[bytes, BytesIO]) -> Dict[str, object]:
    """A decoder to convert response bytes into a dictionary of {filename: value}

    :param data: Data to decode
    :return: A dictionary of decoded files from a tar file
    """
    if isinstance(data, bytes):
        data = BytesIO(data)

    with tarfile.open(fileobj=data) as tar:
        decoded_files = {}
        for member in tar.getmembers():
            file = tar.extractfile(member)
            if file is not None:
                decoded_files[member.name] = decode_data(file.read(), get_data_format(member.name))

        return decoded_files


def decode_sentinelhub_err_msg(response: Response) -> str:
    """Decodes error message from Sentinel Hub service

    :param response: Sentinel Hub service response
    :return: An error message
    """
    if not isinstance(response.content, bytes) or not response.content:
        return ""

    try:
        json_message = json.loads(response.content)
        if "error" in json_message:
            json_message = json_message["error"]

        if isinstance(json_message, str):
            return json_message
        return json.dumps(json_message)
    except JSONDecodeError:
        pass

    try:
        server_message = []
        for elem in decode_data(response.content, MimeType.XML):
            if "ServiceException" in elem.tag or "Message" in elem.tag or elem.tag == "body":
                for text in elem.itertext():
                    stripped_text = text.strip()
                    if stripped_text:
                        server_message.append(stripped_text)
        return " ".join(server_message)
    except ElementTree.ParseError:
        return response.text


def get_jp2_bit_depth(stream: IOBase) -> int:
    """Reads a bit encoding depth of jpeg2000 file in binary stream format

    :param stream: binary stream format
    :return: bit depth
    """
    stream.seek(0)
    while True:
        read_buffer = stream.read(8)
        if len(read_buffer) < 8:
            raise ValueError("Image Header Box not found in JPEG2000 file")

        _, box_id = struct.unpack(">I4s", read_buffer)

        if box_id == b"ihdr":
            read_buffer = stream.read(14)
            params = struct.unpack(">IIHBBBB", read_buffer)
            return (params[3] & 0x7F) + 1


def fix_jp2_image(image: np.ndarray, bit_depth: int) -> np.ndarray:
    """Because Pillow library incorrectly reads JPEG 2000 images with 15-bit encoding this function corrects the
    values in image.

    :param image: image read by opencv library
    :param bit_depth: A bit depth of jp2 image encoding
    :return: corrected image
    """
    if bit_depth in [8, 16]:
        return image
    if bit_depth == 15:
        try:
            return image >> 1
        except TypeError as exception:
            raise IOError(
                "Failed to read JPEG2000 image correctly. Most likely reason is that Pillow did not "
                "install OpenJPEG library correctly. Try reinstalling Pillow from a wheel"
            ) from exception

    raise ValueError(
        f"Bit depth {bit_depth} of jp2 image is currently not supported. Please raise an issue on package Github page"
    )


def get_data_format(filename: str) -> MimeType:
    """Util function to guess format from filename extension

    :param filename: name of file
    :return: file extension
    """
    fmt_ext = filename.split(".")[-1]
    return MimeType.from_string(fmt_ext)



""" Contains information about data collections used by SH """
from dataclasses import dataclass
from enum import Enum
from typing import Tuple

import numpy as np


class Unit(Enum):
    """Collection of all units used by Sentinel Hub"""

    DN = "DN"
    PPB = "PPB"
    INDEX = "INDEX"
    KELVIN = "KELVIN"
    METERS = "METERS"
    MOL_M2 = "MOL_M2"
    KG_M2 = "KG_M2"
    DEGREES = "DEGREES"
    PASCALS = "PASCALS"
    PERCENT = "PERCENT"
    FRACTION = "FRACTION"
    RADIANCE = "RADIANCE"
    KILOMETERS = "KILOMETERS"
    REFLECTANCE = "REFLECTANCE"
    LINEAR_POWER = "LINEAR_POWER"
    OPTICAL_DEPTH = "OPTICAL_DEPTH"
    BRIGHTNESS_TEMPERATURE = "BRIGHTNESS_TEMPERATURE"
    SURFACE_TEMPERATURE = "SURFACE_TEMPERATURE"
    HECTOPASCALS = "HECTOPASCALS"


@dataclass(frozen=True)
class Band:
    """Information about a band available in a collection

    Units and output types need to have the same order i.e. the unit at position 2 has to match the output type
    at position 2. The unit (and output type) at position 0 is considered the default.
    """

    name: str
    units: Tuple[Unit, ...]
    output_types: Tuple[type, ...]


class Bands:
    """
    Different collections of bands taken from `Sentinel Hub docs <https://docs.sentinel-hub.com/api/latest/data/>`__
    """

    SENTINEL1_IW = (
        Band("VV", (Unit.LINEAR_POWER,), (np.float32,)),
        Band("VH", (Unit.LINEAR_POWER,), (np.float32,)),
    )
    SENTINEL1_EW = (
        Band("HH", (Unit.LINEAR_POWER,), (np.float32,)),
        Band("HV", (Unit.LINEAR_POWER,), (np.float32,)),
    )
    SENTINEL1_EW_SH = (Band("HH", (Unit.LINEAR_POWER,), (np.float32,)),)
    SENTINEL2_L1C = tuple(
        Band(name, (Unit.REFLECTANCE, Unit.DN), (np.float32, np.uint16))
        for name in ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B8A", "B09", "B10", "B11", "B12"]
    )
    SENTINEL2_L2A = tuple(
        Band(name, (Unit.REFLECTANCE, Unit.DN), (np.float32, np.uint16))
        for name in ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B8A", "B09", "B11", "B12"]
    )
    SENTINEL3_OLCI = tuple(
        Band(f"B{str(index).zfill(2)}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 22)
    )
    SENTINEL3_SLSTR = (
        *(Band(f"S{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 7)),
        *(Band(name, (Unit.BRIGHTNESS_TEMPERATURE,), (np.float32,)) for name in ["S7", "S8", "S9", "F1", "F2"]),
    )
    SENTINEL5P = (
        *(Band(name, (Unit.DN,), (np.float32,)) for name in ["CO", "HCHO", "NO2", "O3", "SO2"]),
        Band("CH4", (Unit.PPB,), (np.float32,)),
        *(Band(name, (Unit.INDEX,), (np.float32,)) for name in ["AER_AI_340_380", "AER_AI_354_388"]),
        *(Band(name, (Unit.PASCALS,), (np.float32,)) for name in ["CLOUD_BASE_PRESSURE", "CLOUD_TOP_PRESSURE"]),
        *(Band(name, (Unit.METERS,), (np.float32,)) for name in ["CLOUD_BASE_HEIGHT", "CLOUD_TOP_HEIGHT"]),
        Band("CLOUD_OPTICAL_THICKNESS", (Unit.OPTICAL_DEPTH,), (np.float32,)),
        Band("CLOUD_FRACTION", (Unit.FRACTION,), (np.float32,)),
    )
    LANDSAT_OT_L1 = (
        *(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 10)),
        *(Band(f"B{index}", (Unit.BRIGHTNESS_TEMPERATURE,), (np.float32,)) for index in [10, 11]),
    )
    LANDSAT_OT_L2 = (
        *(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 8)),
        Band("B10", (Unit.SURFACE_TEMPERATURE,), (np.float32,)),
    )
    LANDSAT_ETM_L1 = (
        *(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 6)),
        *(Band(f"B06_VCID_{subindex}", (Unit.BRIGHTNESS_TEMPERATURE,), (np.float32,)) for subindex in [1, 2]),
        *(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in [7, 8]),
    )
    LANDSAT_ETM_L2 = (
        *(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 6)),
        Band("B06", (Unit.SURFACE_TEMPERATURE,), (np.float32,)),
        Band("B07", (Unit.REFLECTANCE,), (np.float32,)),
    )
    LANDSAT_TM_L1 = (
        *(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 6)),
        Band("B06", (Unit.BRIGHTNESS_TEMPERATURE,), (np.float32,)),
        Band("B07", (Unit.REFLECTANCE,), (np.float32,)),
    )
    LANDSAT_TM_L2 = (
        *(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 6)),
        Band("B06", (Unit.SURFACE_TEMPERATURE,), (np.float32,)),
        Band("B07", (Unit.REFLECTANCE,), (np.float32,)),
    )
    LANDSAT_MSS_L1 = tuple(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 5))
    MODIS = tuple(Band(f"B0{index}", (Unit.REFLECTANCE,), (np.float32,)) for index in range(1, 8))
    DEM = (Band("DEM", (Unit.METERS,), (np.float32,)),)


class MetaBands:
    """
    Different collections of bands taken from `Sentinel Hub docs <https://docs.sentinel-hub.com/api/latest/data/>`__
    """

    SENTINEL1 = (
        *(Band(name, (Unit.DN,), (np.float32,)) for name in ["localIncidenceAngle", "scatteringArea"]),
        *(Band(name, (Unit.DN,), (bool,)) for name in ["shadowMask", "dataMask"]),
    )
    SENTINEL2_L1C = (
        *(
            Band(name, (Unit.DEGREES,), (np.float32,))
            for name in ["sunAzimuthAngles", "viewAzimuthMean", "sunZenithAngles", "viewZenithMean"]
        ),
        *(Band(name, (Unit.DN,), (np.uint8,)) for name in ["CLP", "CLM"]),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    SENTINEL2_L2A = (
        Band("AOT", (Unit.OPTICAL_DEPTH, Unit.DN), (np.float32, np.uint16)),
        *(
            Band(name, (Unit.DEGREES,), (np.float32,))
            for name in ["sunAzimuthAngles", "viewAzimuthMean", "sunZenithAngles", "viewZenithMean"]
        ),
        *(Band(name, (Unit.PERCENT,), (np.uint8,)) for name in ["SNW", "CLD"]),
        *(Band(name, (Unit.DN,), (np.uint8,)) for name in ["SCL", "CLP", "CLM"]),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    SENTINEL3_OLCI = (
        *(Band(name, (Unit.DEGREES,), (np.float32,)) for name in ["VAA", "VZA", "SAA", "SZA"]),
        Band("HUMIDITY", (Unit.PERCENT,), (np.float32,)),
        Band("SEA_LEVEL_PRESSURE", (Unit.HECTOPASCALS,), (np.float32,)),
        *(Band(name, (Unit.KG_M2,), (np.float32,)) for name in ["TOTAL_COLUMN_OZONE", "TOTAL_COLUMN_WATER_VAPOUR"]),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    SENTINEL3_SLSTR = (Band("dataMask", (Unit.DN,), (bool,)),)
    SENTINEL5P = (Band("dataMask", (Unit.DN,), (bool,)),)
    LANDSAT_OT_L1 = (
        *(Band(name, (Unit.DN,), (np.uint16,)) for name in ["BQA", "QA_RADSAT"]),
        *(Band(name, (Unit.DEGREES,), (np.float32,)) for name in ["VAA", "VZA", "SAA", "SZA"]),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    LANDSAT_OT_L2 = (
        *(Band(name, (Unit.DN,), (np.uint16,)) for name in ["BQA", "QA_RADSAT"]),
        Band("SR_QA_AEROSOL", (Unit.DN,), (np.uint8,)),
        Band("ST_QA", (Unit.KELVIN,), (np.float32,)),
        Band("ST_TRAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_URAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_DRAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_ATRAN", (Unit.FRACTION,), (np.float32,)),
        Band("ST_EMIS", (Unit.FRACTION,), (np.float32,)),
        Band("ST_EMSD", (Unit.FRACTION,), (np.float32,)),
        Band("ST_CDIST", (Unit.KILOMETERS,), (np.float32,)),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    LANDSAT_ETM_L1 = (
        *(Band(name, (Unit.DN,), (np.uint16,)) for name in ["BQA", "QA_RADSAT"]),
        *(Band(name, (Unit.DEGREES,), (np.float32,)) for name in ["VAA", "VZA", "SAA", "SZA"]),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    LANDSAT_ETM_L2 = (
        *(Band(name, (Unit.DN,), (np.uint16,)) for name in ["BQA", "QA_RADSAT"]),
        Band("ST_TRAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_URAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_DRAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_ATRAN", (Unit.FRACTION,), (np.float32,)),
        Band("ST_EMIS", (Unit.FRACTION,), (np.float32,)),
        Band("ST_EMSD", (Unit.FRACTION,), (np.float32,)),
        Band("ST_CDIST", (Unit.KILOMETERS,), (np.float32,)),
        Band("SR_ATMOS_OPACITY", (Unit.FRACTION,), (np.float32,)),
        Band("SR_CLOUD_QA", (Unit.DN,), (np.float32,)),
        Band("ST_QA", (Unit.KELVIN,), (np.float32,)),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    LANDSAT_TM_L1 = (
        *(Band(name, (Unit.DN,), (np.uint16,)) for name in ["BQA", "QA_RADSAT"]),
        *(Band(name, (Unit.DEGREES,), (np.float32,)) for name in ["VAA", "VZA", "SAA", "SZA"]),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    LANDSAT_TM_L2 = (
        *(Band(name, (Unit.DN,), (np.uint16,)) for name in ["BQA", "QA_RADSAT"]),
        Band("ST_TRAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_URAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_DRAD", (Unit.RADIANCE,), (np.float32,)),
        Band("ST_ATRAN", (Unit.FRACTION,), (np.float32,)),
        Band("ST_EMIS", (Unit.FRACTION,), (np.float32,)),
        Band("ST_EMSD", (Unit.FRACTION,), (np.float32,)),
        Band("ST_CDIST", (Unit.KILOMETERS,), (np.float32,)),
        Band("SR_ATMOS_OPACITY", (Unit.FRACTION,), (np.float32,)),
        Band("SR_CLOUD_QA", (Unit.DN,), (np.float32,)),
        Band("ST_QA", (Unit.KELVIN,), (np.float32,)),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    LANDSAT_MSS_L1 = (
        *(Band(name, (Unit.DN,), (np.uint16,)) for name in ["BQA", "QA_RADSAT"]),
        Band("dataMask", (Unit.DN,), (bool,)),
    )
    MODIS = (Band("dataMask", (Unit.DN,), (bool,)),)
    DEM = (Band("dataMask", (Unit.DN,), (bool,)),)



"""
The part of the package that implements interface with Sentinel Hub services.
"""

from .batch import (
    BatchCollection,
    BatchRequest,
    BatchRequestStatus,
    BatchTileStatus,
    BatchUserAction,
    SentinelHubBatch,
    monitor_batch_analysis,
    monitor_batch_job,
)
from .byoc import ByocCollection, ByocCollectionAdditionalData, ByocCollectionBand, ByocTile, SentinelHubBYOC
from .catalog import SentinelHubCatalog
from .fis import FisRequest
from .ogc import WcsRequest, WmsRequest
from .process import SentinelHubRequest
from .statistical import SentinelHubStatistical
from .wfs import WebFeatureService



"""
Constants related to AWS functionalities of the package
"""
import itertools as it
from enum import Enum

from ..constants import MimeType


class AwsConstants:
    """Initialisation of every constant used by AWS classes

    For each supported data collection it contains lists of all possible bands and all possible metadata files:

        - S2_L1C_BANDS and S2_L1C_METAFILES
        - S2_L2A_BANDS and S2_L2A_METAFILES

    It also contains dictionary of all possible files and their formats: AWS_FILES
    """

    # General constants:
    SOURCE_ID_LIST = ["L1C", "L2A"]
    TILE_INFO = "tileInfo"
    PRODUCT_INFO = "productInfo"
    METADATA = "metadata"
    PREVIEW = "preview"
    PREVIEW_JP2 = "preview*"
    QI_LIST = ["DEFECT", "DETFOO", "NODATA", "SATURA", "TECQUA"]
    QI_LIST_V4 = ["DETFOO", "QUALIT"]  # new filenames since Processing Baseline 04.00
    QI_MSK_CLASSI = "qi/CLASSI_B00"
    QI_MSK_CLASSI_V4 = "qi/MSK_CLASSI_B00"  # new filename since Processing Baseline 04.00
    QI_MSK_CLOUD = "qi/MSK_CLOUDS_B00"
    AUX_DATA = "AUX_DATA"
    DATASTRIP = "DATASTRIP"
    GRANULE = "GRANULE"
    HTML = "HTML"
    INFO = "rep_info"
    QI_DATA = "QI_DATA"
    IMG_DATA = "IMG_DATA"
    INSPIRE = "inspire"
    MANIFEST = "manifest"
    TCI = "TCI"
    PVI = "PVI"
    CAMSFO = "auxiliary/CAMSFO"
    AUX_CAMSFO = "auxiliary/AUX_CAMSFO"
    ECMWFT = "auxiliary/ECMWFT"
    AUX_ECMWFT = "auxiliary/AUX_ECMWFT"
    DATASTRIP_METADATA = "datastrip/*/metadata"

    # More constants about L2A
    AOT = "AOT"
    WVP = "WVP"
    SCL = "SCL"
    VIS = "VIS"
    L2A_MANIFEST = "L2AManifest"
    REPORT = "report"
    GIPP = "auxiliary/GIP_TL"
    FORMAT_CORRECTNESS = "FORMAT_CORRECTNESS"
    GENERAL_QUALITY = "GENERAL_QUALITY"
    GEOMETRIC_QUALITY = "GEOMETRIC_QUALITY"
    RADIOMETRIC_QUALITY = "RADIOMETRIC_QUALITY"
    SENSOR_QUALITY = "SENSOR_QUALITY"
    L2A_QUALITY = "L2A_QUALITY"
    QUALITY_REPORTS = [FORMAT_CORRECTNESS, GENERAL_QUALITY, GEOMETRIC_QUALITY, RADIOMETRIC_QUALITY, SENSOR_QUALITY]
    CLASS_MASKS = ["SNW", "CLD"]
    R10m = "R10m"
    R20m = "R20m"
    R60m = "R60m"
    RESOLUTIONS = [R10m, R20m, R60m]
    S2_L2A_BAND_MAP = {
        R10m: ["B02", "B03", "B04", "B08", AOT, TCI, WVP],
        R20m: ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B8A", "B11", "B12", AOT, SCL, TCI, VIS, WVP],
        R60m: ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B8A", "B09", "B11", "B12", AOT, SCL, TCI, WVP],
    }

    # Order of elements in following lists is important
    # Sentinel-2 L1C products:
    S2_L1C_BANDS = ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B8A", "B09", "B10", "B11", "B12"]
    S2_L1C_METAFILES = (
        [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, DATASTRIP_METADATA]
        + [f"datastrip/*/qi/{qi_report}_report" for qi_report in QUALITY_REPORTS]
        + [PREVIEW, PREVIEW_JP2, TCI]
        + [f"{preview}/{band}" for preview, band in it.zip_longest([], S2_L1C_BANDS, fillvalue=PREVIEW)]
        + [QI_MSK_CLASSI, QI_MSK_CLASSI_V4, QI_MSK_CLOUD]
        + [f"qi/MSK_{qi}_{band}" for qi, band in it.product(QI_LIST, S2_L1C_BANDS)]
        + [f"qi/{qi}_{band}" for qi, band in it.product(QI_LIST_V4, S2_L1C_BANDS)]
        + [f"qi/{qi_report}" for qi_report in [FORMAT_CORRECTNESS, GENERAL_QUALITY, GEOMETRIC_QUALITY, SENSOR_QUALITY]]
        + [CAMSFO, ECMWFT]
    )

    # Sentinel-2 L2A products:
    S2_L2A_BANDS = [
        f"{resolution}/{band}" for resolution, band_list in sorted(S2_L2A_BAND_MAP.items()) for band in band_list
    ]
    S2_L2A_METAFILES = (
        [PRODUCT_INFO, TILE_INFO, METADATA, INSPIRE, MANIFEST, L2A_MANIFEST, REPORT, DATASTRIP_METADATA]
        + [f"datastrip/*/qi/{qi_report}" for qi_report in QUALITY_REPORTS]
        + [f"qi/{source_id}_PVI" for source_id in SOURCE_ID_LIST]
        + [f'qi/{mask}_{res.lstrip("R")}' for mask, res in it.product(CLASS_MASKS, [R20m, R60m])]
        + [f"qi/MSK_{qi}_{band}" for qi, band in it.product(QI_LIST, S2_L1C_BANDS)]
        + [f"qi/{qi}_{band}" for qi, band in it.product(QI_LIST_V4, S2_L1C_BANDS)]
        + [QI_MSK_CLASSI, QI_MSK_CLASSI_V4, QI_MSK_CLOUD]
        + [f"qi/{qi_report}" for qi_report in QUALITY_REPORTS + [L2A_QUALITY]]
        + [AUX_CAMSFO, ECMWFT, AUX_ECMWFT, GIPP]
    )

    # Product files with formats:
    PRODUCT_FILES = {
        **{
            PRODUCT_INFO: MimeType.JSON,
            METADATA: MimeType.XML,
            INSPIRE: MimeType.XML,
            MANIFEST: MimeType.SAFE,
            L2A_MANIFEST: MimeType.XML,
            REPORT: MimeType.XML,
            DATASTRIP_METADATA: MimeType.XML,
        },
        **{f"datastrip/*/qi/{qi_report}": MimeType.XML for qi_report in QUALITY_REPORTS},
    }
    # Tile files with formats:
    TILE_FILES = {
        **{
            TILE_INFO: MimeType.JSON,
            PRODUCT_INFO: MimeType.JSON,
            METADATA: MimeType.XML,
            PREVIEW: MimeType.JPG,
            PREVIEW_JP2: MimeType.JP2,
            TCI: MimeType.JP2,
            QI_MSK_CLASSI: MimeType.JP2,
            QI_MSK_CLASSI_V4: MimeType.JP2,
            QI_MSK_CLOUD: MimeType.GML,
            CAMSFO: MimeType.RAW,
            AUX_CAMSFO: MimeType.RAW,
            ECMWFT: MimeType.RAW,
            AUX_ECMWFT: MimeType.RAW,
            GIPP: MimeType.XML,
        },
        **{f"qi/{qi_report}": MimeType.XML for qi_report in QUALITY_REPORTS + [L2A_QUALITY]},
        **{f"{preview}/{band}": MimeType.JP2 for preview, band in it.zip_longest([], S2_L1C_BANDS, fillvalue=PREVIEW)},
        **{f"qi/MSK_{qi}_{band}": MimeType.GML for qi, band in it.product(QI_LIST, S2_L1C_BANDS)},
        **{band: MimeType.JP2 for band in S2_L1C_BANDS},
        **{band: MimeType.JP2 for band in S2_L2A_BANDS},
        **{f"qi/{source_id}_PVI": MimeType.JP2 for source_id in SOURCE_ID_LIST},
        **{f'qi/{mask}_{res.lstrip("R")}': MimeType.JP2 for mask, res in it.product(CLASS_MASKS, [R20m, R60m])},
    }

    # All files joined together
    AWS_FILES = {
        **PRODUCT_FILES,
        **{filename.split("/")[-1]: data_format for filename, data_format in PRODUCT_FILES.items()},
        **TILE_FILES,
        **{filename.split("/")[-1]: data_format for filename, data_format in TILE_FILES.items()},
    }


class EsaSafeType(Enum):
    """Enum constants class for ESA .SAFE type.

    Types are OLD_TYPE and COMPACT_TYPE
    """

    OLD_TYPE = "old_type"
    COMPACT_TYPE = "compact_type"



"""
Implementation of `Sentinel Hub Process API interface <https://docs.sentinel-hub.com/api/latest/api/process/>`__.
"""
from typing import Any, Dict, List, Optional, Tuple, Union

from ..constants import MimeType
from ..download import SentinelHubDownloadClient
from ..geometry import BBox, Geometry
from ..type_utils import JsonDict
from .base_request import InputDataDict, SentinelHubBaseApiRequest
from .utils import _update_other_args


class SentinelHubRequest(SentinelHubBaseApiRequest):
    """Sentinel Hub Process API interface

    For more information check
    `Process API documentation <https://docs.sentinel-hub.com/api/latest/api/process/>`__.
    """

    _SERVICE_ENDPOINT = "process"

    def __init__(
        self,
        evalscript: str,
        input_data: List[Union[JsonDict, InputDataDict]],
        responses: List[JsonDict],
        bbox: Optional[BBox] = None,
        geometry: Optional[Geometry] = None,
        size: Optional[Tuple[int, int]] = None,
        resolution: Optional[Tuple[float, float]] = None,
        **kwargs: Any
    ):
        """
        For details of certain parameters check the
        `Process API reference <https://docs.sentinel-hub.com/api/latest/reference/#operation/process>`_.

        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :param input_data: A list of input dictionary objects as described in the API reference. It can be generated
            with the helper method `SentinelHubRequest.input_data`
        :param responses: A list of `output.responses` objects as described in the API reference. It can be generated
            with the helper function `SentinelHubRequest.output_response`
        :param bbox: Bounding box describing the area of interest.
        :param geometry: Geometry describing the area of interest.
        :param size: Size of the image.
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        if not isinstance(evalscript, str):
            raise ValueError("'evalscript' should be a string")

        parsed_mime_type = MimeType.from_string(responses[0]["format"]["type"])
        self._mime_type = MimeType.TAR if len(responses) > 1 else parsed_mime_type

        self.payload = self.body(
            request_bounds=self.bounds(bbox=bbox, geometry=geometry),
            request_data=input_data,
            request_output=self.output(size=size, resolution=resolution, responses=responses),
            evalscript=evalscript,
        )

        super().__init__(SentinelHubDownloadClient, **kwargs)

    @property
    def mime_type(self) -> MimeType:
        return self._mime_type

    @staticmethod
    def body(
        request_bounds: JsonDict,
        request_data: List[JsonDict],
        evalscript: str,
        request_output: Optional[JsonDict] = None,
        other_args: Optional[JsonDict] = None,
    ) -> JsonDict:
        """Generate the Process API request body

        :param request_bounds: A dictionary as generated by `SentinelHubRequest.bounds` helper method.
        :param request_data: A list of dictionaries as generated by `SentinelHubRequest.input_data` helper method.
        :param evalscript: `Evalscript <https://docs.sentinel-hub.com/api/latest/#/Evalscript/>`__.
        :param request_output: A dictionary as generated by `SentinelHubRequest.output` helper method.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        request_body = {"input": {"bounds": request_bounds, "data": request_data}, "evalscript": evalscript}

        if request_output is not None:
            request_body["output"] = request_output

        if other_args:
            _update_other_args(request_body, other_args)

        return request_body

    @staticmethod
    def output_response(
        identifier: str, response_format: Union[str, MimeType], other_args: Optional[JsonDict] = None
    ) -> JsonDict:
        """Generate an element of `output.responses` as described in the Process API reference.

        :param identifier: Identifier of the output response.
        :param response_format: A mime type of one of 'png', 'json', 'jpeg', 'tiff'.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        output_response = {"identifier": identifier, "format": {"type": MimeType(response_format).get_string()}}

        if other_args:
            _update_other_args(output_response, other_args)

        return output_response

    @staticmethod
    def output(
        responses: List[JsonDict],
        size: Optional[Tuple[int, int]] = None,
        resolution: Optional[Tuple[float, float]] = None,
        other_args: Optional[Dict] = None,
    ) -> JsonDict:
        """Generate an `output` part of the request as described in the Process API reference

        :param responses: A list of objects in `output.responses` as generated by `SentinelHubRequest.output_response`.
        :param size: Size of the image.
        :param resolution: Resolution of the image. It has to be in units compatible with the given CRS.
        :param other_args: Additional dictionary of arguments. If provided, the resulting dictionary will get updated
            by it.
        """
        if size and resolution:
            raise ValueError("Either size or resolution argument should be given, not both.")

        request_output: JsonDict = {"responses": responses}

        if size:
            request_output["width"], request_output["height"] = size
        if resolution:
            request_output["resx"], request_output["resy"] = resolution

        if other_args:
            _update_other_args(request_output, other_args)

        return request_output



"""
This module lists all externally useful classes and functions
"""

from ._version import __version__
from .api import (
    BatchCollection,
    BatchRequest,
    BatchRequestStatus,
    BatchTileStatus,
    BatchUserAction,
    ByocCollection,
    ByocCollectionAdditionalData,
    ByocCollectionBand,
    ByocTile,
    FisRequest,
    SentinelHubBatch,
    SentinelHubBYOC,
    SentinelHubCatalog,
    SentinelHubRequest,
    SentinelHubStatistical,
    WcsRequest,
    WebFeatureService,
    WmsRequest,
    monitor_batch_analysis,
    monitor_batch_job,
    opensearch,
)
from .api.opensearch import get_area_dates, get_area_info, get_tile_info, get_tile_info_id
from .areas import (
    BatchSplitter,
    BBoxSplitter,
    CustomGridSplitter,
    OsmSplitter,
    TileSplitter,
    UtmGridSplitter,
    UtmZoneSplitter,
)
from .config import SHConfig
from .constants import (
    CRS,
    CustomUrlParam,
    HistogramType,
    MimeType,
    MosaickingOrder,
    ResamplingType,
    ServiceType,
    ServiceUrl,
    SHConstants,
)
from .data_collections import DataCollection, DataSource
from .data_collections_bands import Band, Unit
from .download import (
    DownloadClient,
    DownloadRequest,
    SentinelHubDownloadClient,
    SentinelHubSession,
    SentinelHubStatisticalDownloadClient,
)
from .exceptions import AwsDownloadFailedException, DownloadFailedException
from .geo_utils import (
    bbox_to_dimensions,
    bbox_to_resolution,
    get_image_dimension,
    get_utm_bbox,
    get_utm_crs,
    pixel_to_utm,
    to_utm_bbox,
    to_wgs84,
    transform_point,
    utm_to_pixel,
    wgs84_to_pixel,
    wgs84_to_utm,
)
from .geometry import BBox, BBoxCollection, Geometry
from .geopedia import GeopediaFeatureIterator, GeopediaImageRequest, GeopediaSession, GeopediaWmsRequest
from .io_utils import read_data, write_data
from .os_utils import create_parent_folder, get_content_list, get_file_list, get_folder_list, make_folder, rename, size
from .time_utils import filter_times, is_valid_time, parse_time, parse_time_interval, serialize_time



"""
Module for managing configuration data from `config.json`
"""

import copy
import json
import numbers
import os
from typing import Any, Dict, Iterable, List, Optional, Union

ConfigDict = Dict[str, Union[str, int, float]]


class SHConfig:  # pylint: disable=too-many-instance-attributes
    """A sentinelhub-py package configuration class.

    The class reads during its first initialization the configurable settings from ``./config.json`` file:

        - `instance_id`: An instance ID for Sentinel Hub service used for OGC requests.
        - `sh_client_id`: User's OAuth client ID for Sentinel Hub service
        - `sh_client_secret`: User's OAuth client secret for Sentinel Hub service
        - `sh_base_url`: There exist multiple deployed instances of Sentinel Hub service, this parameter defines the
          location of a specific service instance.
        - `sh_auth_base_url`: Base url for Sentinel Hub Authentication service. Authentication is typically sent to the
          main service deployment even if `sh_base_url` points to another deployment.
        - `geopedia_wms_url`: Base url for Geopedia WMS services.
        - `geopedia_rest_url`: Base url for Geopedia REST services.
        - `aws_access_key_id`: Access key for AWS Requester Pays buckets.
        - `aws_secret_access_key`: Secret access key for AWS Requester Pays buckets.
        - `aws_session_token`: A session token for your AWS account. It is only needed when you are using temporary
          credentials.
        - `aws_metadata_url`: Base url for publicly available metadata files
        - `aws_s3_l1c_bucket`: Name of Sentinel-2 L1C bucket at AWS s3 service.
        - `aws_s3_l2a_bucket`: Name of Sentinel-2 L2A bucket at AWS s3 service.
        - `opensearch_url`: Base url for Sentinelhub Opensearch service.
        - `max_wfs_records_per_query`: Maximum number of records returned for each WFS query.
        - `max_opensearch_records_per_query`: Maximum number of records returned for each Opensearch query.
        - `max_download_attempts`: Maximum number of download attempts from a single URL until an error will be raised.
        - `download_sleep_time`: Number of seconds to sleep between the first failed attempt and the next. Every next
          attempt this number exponentially increases with factor `3`.
        - `download_timeout_seconds`: Maximum number of seconds before download attempt is canceled.
        - `number_of_download_processes`: Number of download processes, used to calculate rate-limit sleep time.

    Usage in the code:

        * ``SHConfig().sh_base_url``
        * ``SHConfig().instance_id``

    """

    CREDENTIALS = {
        "instance_id",
        "sh_client_id",
        "sh_client_secret",
        "aws_access_key_id",
        "aws_secret_access_key",
        "aws_session_token",
    }
    CONFIG_PARAMS = [
        "instance_id",
        "sh_client_id",
        "sh_client_secret",
        "sh_base_url",
        "sh_auth_base_url",
        "geopedia_wms_url",
        "geopedia_rest_url",
        "aws_access_key_id",
        "aws_secret_access_key",
        "aws_session_token",
        "aws_metadata_url",
        "aws_s3_l1c_bucket",
        "aws_s3_l2a_bucket",
        "opensearch_url",
        "max_wfs_records_per_query",
        "max_opensearch_records_per_query",
        "max_download_attempts",
        "download_sleep_time",
        "download_timeout_seconds",
        "number_of_download_processes",
    ]

    _cache: Optional[Dict[str, Any]] = None

    def __init__(self, hide_credentials: bool = False, use_defaults: bool = False):
        """
        :param hide_credentials: If `True` then methods that provide the entire content of the config object will mask
            out all credentials. But credentials could still be accessed directly from config object attributes. The
            default is `False`.
        :param use_defaults: Does not load the configuration file, returns config object with defaults only.
        """

        self.instance_id: str = ""
        self.sh_client_id: str = ""
        self.sh_client_secret: str = ""
        self.sh_base_url: str = "https://services.sentinel-hub.com"
        self.sh_auth_base_url: str = "https://services.sentinel-hub.com"
        self.geopedia_wms_url: str = "https://service.geopedia.world"
        self.geopedia_rest_url: str = "https://www.geopedia.world/rest"
        self.aws_access_key_id: str = ""
        self.aws_secret_access_key: str = ""
        self.aws_session_token: str = ""
        self.aws_metadata_url: str = "https://roda.sentinel-hub.com"
        self.aws_s3_l1c_bucket: str = "sentinel-s2-l1c"
        self.aws_s3_l2a_bucket: str = "sentinel-s2-l2a"
        self.opensearch_url: str = "http://opensearch.sentinel-hub.com/resto/api/collections/Sentinel2"
        self.max_wfs_records_per_query: int = 100
        self.max_opensearch_records_per_query: int = 500  # pylint: disable=invalid-name
        self.max_download_attempts: int = 4
        self.download_sleep_time: float = 5.0
        self.download_timeout_seconds: float = 120.0
        self.number_of_download_processes: int = 1

        self._hide_credentials = hide_credentials

        if not use_defaults:
            for param, value in self._global_cache.items():
                setattr(self, param, value)

    def _validate_values(self) -> None:
        """Ensures that the values are aligned with expectations."""
        default = SHConfig(use_defaults=True)
        for param in self.CONFIG_PARAMS:
            value = getattr(self, param)
            default_value = getattr(default, param)
            param_type = type(default_value)

            if isinstance(value, str) and value.startswith("http"):
                value = value.rstrip("/")
            if (param_type is float) and isinstance(value, numbers.Number):
                continue
            if not isinstance(value, param_type):
                raise ValueError(f"Value of parameter '{param}' must be of type {param_type.__name__}")
        if self.max_wfs_records_per_query > 100:
            raise ValueError("Value of config parameter 'max_wfs_records_per_query' must be at most 100")
        if self.max_opensearch_records_per_query > 500:
            raise ValueError("Value of config parameter 'max_opensearch_records_per_query' must be at most 500")

    def __getitem__(self, name: str) -> Union[str, int, float]:
        """Config parameters can also be accessed as items."""
        if name in self.CONFIG_PARAMS:
            return getattr(self, name)
        raise KeyError(f"'{name}' is not a supported config parameter")

    def __str__(self) -> str:
        """Content of SHConfig in json schema. If `hide_credentials` is set to `True` then credentials will be
        masked.
        """
        return json.dumps(self.get_config_dict(), indent=2)

    def __repr__(self) -> str:
        """Representation of SHConfig parameters. If `hide_credentials` is set to `True` then credentials will be
        masked.
        """
        repr_list = [f"{self.__class__.__name__}("]

        for key, value in self.get_config_dict().items():
            repr_list.append(f"{key}={repr(value)},")

        return "\n  ".join(repr_list).strip(",") + "\n)"

    @property
    def _global_cache(self) -> Dict[str, Any]:
        """Uses a class attribute to store a global instance of a class with config parameters."""
        if SHConfig._cache is None:
            loaded_instance = SHConfig.load(self.get_config_location())
            SHConfig._cache = {param: getattr(loaded_instance, param) for param in SHConfig.CONFIG_PARAMS}
        return SHConfig._cache

    @classmethod
    def load(cls, filename: str) -> "SHConfig":
        """Method that loads configuration parameters from a file. Does not affect global settings.

        :param filename: Path to file from which to read configuration.
        """
        with open(filename, "r") as cfg_file:
            config_dict = json.load(cfg_file)

        config = cls(use_defaults=True)
        for param, value in config_dict.items():
            if param in cls.CONFIG_PARAMS:
                setattr(config, param, value)

        config._validate_values()
        return config

    def save(self, filename: Optional[str] = None) -> None:
        """Method that saves configuration parameter changes from instance of SHConfig class to global config class and
        to `config.json` file.

        :param filename: Optional name of file to which to save configuration. If not specified saves to global default.

        :Example:
            ``my_config = SHConfig()`` \n
            ``my_config.instance_id = '<new instance id>'`` \n
            ``my_config.save()``
        """
        self._validate_values()

        is_changed = False
        for param in self.CONFIG_PARAMS:
            if getattr(self, param) != self._global_cache[param]:
                is_changed = True
                self._global_cache[param] = getattr(self, param)  # pylint: disable=unsupported-assignment-operation

        if is_changed:
            config_dict = {param: getattr(self, param) for param in self.CONFIG_PARAMS}
            with open(filename or self.get_config_location(), "w") as cfg_file:
                json.dump(config_dict, cfg_file, indent=2)

    def copy(self) -> "SHConfig":
        """Makes a copy of an instance of `SHConfig`"""
        return copy.copy(self)

    def reset(self, params: Union[str, Iterable[str], object] = ...) -> None:
        """Resets configuration class to initial values. Use `SHConfig.save()` method in order to save this change.

        :param params: Parameters which will be reset. Parameters can be specified with a list of names, e.g.
            ``['instance_id', 'aws_access_key_id', 'aws_secret_access_key']``, or as a single name, e.g.
            ``'sh_base_url'``. By default, all parameters will be reset and default value is ``Ellipsis``.
        """
        default = SHConfig(use_defaults=True)

        if params is ...:
            params = self.get_params()
        if isinstance(params, str):
            self._reset_param(params, default)
        elif isinstance(params, Iterable):
            for param in params:
                self._reset_param(param, default)
        else:
            raise ValueError(
                f"Parameters must be specified in form of a list of strings or as a single string, instead got {params}"
            )

    def _reset_param(self, param: str, default: "SHConfig") -> None:
        """Resets a single parameter

        :param param: A configuration parameter
        """
        if param not in self.get_params():
            raise ValueError(f"Cannot reset unknown parameter '{param}'")
        setattr(self, param, getattr(default, param))

    def get_params(self) -> List[str]:
        """Returns a list of parameter names

        :return: List of parameter names
        """
        return list(self.CONFIG_PARAMS)

    def get_config_dict(self) -> ConfigDict:
        """Get a dictionary representation of `SHConfig` class. If `hide_credentials` is set to `True` then
        credentials will be masked.

        :return: A dictionary with configuration parameters
        """
        config_params = {param: getattr(self, param) for param in self.CONFIG_PARAMS}

        if self._hide_credentials:
            config_params = {param: self._mask_credentials(param, value) for param, value in config_params.items()}

        return config_params

    @classmethod
    def get_config_location(cls) -> str:
        """Returns location of configuration file on disk

        :return: File path of `config.json` file
        """
        config_file = os.path.join(os.path.dirname(__file__), "config.json")

        if not os.path.isfile(config_file):
            with open(config_file, "w") as cfg_file:
                default_dict = cls(use_defaults=True).get_config_dict()
                json.dump(default_dict, cfg_file, indent=2)

        return config_file

    def _mask_credentials(self, param: str, value: object) -> object:
        """In case a parameter that holds credentials is given it will mask its value"""
        if not (param in self.CREDENTIALS and value):
            return value
        if not isinstance(value, str):
            raise ValueError(f"Parameter '{param}' should be a string but {value} found")

        hide_size = min(max(len(value) - 4, 10), len(value))
        return "*" * hide_size + value[hide_size:]

    def get_sh_oauth_url(self) -> str:
        """Provides URL for Sentinel Hub authentication endpoint

        :return: A URL endpoint
        """
        return f"{self.sh_auth_base_url}/oauth/token"

    def get_sh_process_api_url(self) -> str:
        """Provides URL for Sentinel Hub Process API endpoint

        :return: A URL endpoint
        """
        return f"{self.sh_base_url}/api/v1/process"

    def get_sh_ogc_url(self) -> str:
        """Provides URL for Sentinel Hub OGC endpoint

        :return: A URL endpoint
        """
        return f"{self.sh_base_url}/ogc"

    def get_sh_rate_limit_url(self) -> str:
        """Provides URL for Sentinel Hub rate limiting endpoint

        :return: A URL endpoint
        """
        return f"{self.sh_auth_base_url}/aux/ratelimit"

    def raise_for_missing_instance_id(self) -> None:
        """In case Sentinel Hub instance ID is missing it raises an informative error

        :raises: ValueError
        """
        if not self.instance_id:
            raise ValueError(
                "Sentinel Hub instance ID is missing. "
                "Either provide it with SHConfig object or save it into config.json configuration file. "
                "Check https://sentinelhub-py.readthedocs.io/en/latest/configure.html for more info."
            )



"""
Implementation of base interface classes of this package.
"""
import copy
import os
from abc import ABCMeta, abstractmethod
from typing import Generic, Iterable, List, Optional, Tuple, TypeVar

import numpy as np

from .config import SHConfig
from .download import DownloadClient, DownloadRequest
from .os_utils import make_folder
from .type_utils import JsonDict

_T = TypeVar("_T")


class DataRequest(metaclass=ABCMeta):
    """A base abstract class for all implementations of data request interface.

    Every data request type can write the fetched data to disk and then read it again (and hence avoid the need to
    download the same data again).
    """

    def __init__(
        self, download_client_class: type, *, data_folder: Optional[str] = None, config: Optional[SHConfig] = None
    ):
        """
        :param download_client_class: A class implementing a download client
        :param data_folder: location of the directory where the fetched data will be saved.
        :param config: A custom instance of config class to override parameters from the saved configuration.
        """
        self.download_client_class = download_client_class
        self.data_folder = data_folder
        self.config = config or SHConfig()

        self.download_list: List[DownloadRequest] = []
        self.folder_list: List[str] = []
        self.create_request()

    @abstractmethod
    def create_request(self) -> None:
        """An abstract method for logic of creating download requests"""

    def get_download_list(self) -> List[DownloadRequest]:
        """
        Returns a list of download requests for requested data.

        :return: List of data to be downloaded
        """
        return self.download_list

    def get_filename_list(self) -> List[str]:
        """Returns a list of file names (or paths relative to `data_folder`) where the requested data will be saved
        or read from, if it has already been downloaded and saved.

        :return: A list of filenames
        """
        return [request.get_relative_paths()[1] for request in self.download_list]

    def get_url_list(self) -> List[Optional[str]]:
        """
        Returns a list of urls for requested data.

        :return: List of URLs from where data will be downloaded.
        """
        return [request.url for request in self.download_list]

    def is_valid_request(self) -> bool:
        """Checks if initialized class instance successfully prepared a list of items to download

        :return: `True` if request is valid and `False` otherwise
        """
        return isinstance(self.download_list, list) and all(
            isinstance(request, DownloadRequest) for request in self.download_list
        )

    def get_data(
        self,
        *,
        save_data: bool = False,
        redownload: bool = False,
        data_filter: Optional[List[int]] = None,
        max_threads: Optional[int] = None,
        decode_data: bool = True,
        raise_download_errors: bool = True,
    ) -> List[np.ndarray]:
        """Get requested data either by downloading it or by reading it from the disk (if it
        was previously downloaded and saved).

        :param save_data: flag to turn on/off saving of data to disk. Default is `False`.
        :param redownload: if `True`, download again the requested data even though it's already saved to disk.
            Default is `False`, do not download if data is already available on disk.
        :param data_filter: Used to specify which items will be returned by the method and in which order. E.g. with
            ``data_filter=[0, 2, -1]`` the method will return only 1st, 3rd and last item. Default filter is `None`.
        :param max_threads: Maximum number of threads to be used for download in parallel. The default is
            `max_threads=None` which will use the number of processors on the system multiplied by 5.
        :param decode_data: If `True` (default) it decodes data (e.g., returns image as an array of numbers);
            if `False` it returns binary data.
        :param raise_download_errors: If `True` any error in download process should be raised as
            ``DownloadFailedException``. If `False` failed downloads will only raise warnings and the method will
            return list with `None` values in places where the results of failed download requests should be.
        :return: requested images as numpy arrays, where each array corresponds to a single acquisition and has
            shape ``[height, width, channels]``.
        """
        self._preprocess_request(save_data, True)
        return self._execute_data_download(
            data_filter, redownload, max_threads, raise_download_errors, decode_data=decode_data
        )

    def save_data(
        self,
        *,
        data_filter: Optional[List[int]] = None,
        redownload: bool = False,
        max_threads: Optional[int] = None,
        raise_download_errors: bool = False,
    ) -> None:
        """Saves data to disk. If ``redownload=True`` then the data is redownloaded using ``max_threads`` workers.

        :param data_filter: Used to specify which items will be returned by the method and in which order. E.g. with
            `data_filter=[0, 2, -1]` the method will return only 1st, 3rd and last item. Default filter is `None`.
        :param redownload: data is redownloaded if ``redownload=True``. Default is `False`
        :param max_threads: Maximum number of threads to be used for download in parallel. The default is
            `max_threads=None` which will use the number of processors on the system multiplied by 5.
        :param raise_download_errors: If `True` any error in download process should be raised as
            ``DownloadFailedException``. If `False` failed downloads will only raise warnings.
        """
        self._preprocess_request(True, False)
        self._execute_data_download(data_filter, redownload, max_threads, raise_download_errors)

    def _execute_data_download(
        self,
        data_filter: Optional[List[int]] = None,
        redownload: bool = False,
        max_threads: Optional[int] = None,
        raise_download_errors: bool = False,
        decode_data: bool = True,
    ) -> List[np.ndarray]:
        """Calls download module and executes the download process

        :param data_filter: Used to specify which items will be returned by the method and in which order. E.g. with
            `data_filter=[0, 2, -1]` the method will return only 1st, 3rd and last item. Default filter is `None`.
        :param redownload: data is redownloaded if ``redownload=True``. Default is `False`
        :param max_threads: Maximum number of threads to be used for download in parallel. The default is
            `max_threads=None` which will use the number of processors on the system multiplied by 5.
        :param raise_download_errors: If `True` any error in download process should be raised as
            ``DownloadFailedException``. If `False` failed downloads will only raise warnings.
        :param decode_data: If `True` (default) it decodes data (e.g., returns image as an array of numbers);
            if `False` it returns binary data.
        :return: List of data obtained from download
        """
        is_repeating_filter = False
        if data_filter is None:
            filtered_download_list = self.download_list
        elif isinstance(data_filter, (list, tuple)):
            try:
                filtered_download_list = [self.download_list[index] for index in data_filter]
            except IndexError as exception:
                raise IndexError("Indices of data_filter are out of range") from exception

            filtered_download_list, mapping_list = self._filter_repeating_items(filtered_download_list)
            is_repeating_filter = len(filtered_download_list) < len(mapping_list)
        else:
            raise ValueError("data_filter parameter must be a list of indices")

        client = self.download_client_class(
            redownload=redownload, raise_download_errors=raise_download_errors, config=self.config
        )
        data_list = client.download(filtered_download_list, max_threads=max_threads, decode_data=decode_data)

        if is_repeating_filter:
            data_list = [copy.deepcopy(data_list[index]) for index in mapping_list]

        return data_list

    @staticmethod
    def _filter_repeating_items(download_list: List[DownloadRequest]) -> Tuple[List[DownloadRequest], List[int]]:
        """Because of data_filter some requests in download list might be the same. In order not to download them again
        this method will reduce the list of requests. It will also return a mapping list which can be used to
        reconstruct the previous list of download requests.

        :param download_list: List of download requests
        :return: reduced download list with unique requests and mapping list
        """
        unique_requests_map = {}
        mapping_list = []
        unique_download_list: List[DownloadRequest] = []
        for download_request in download_list:
            if download_request not in unique_requests_map:
                unique_requests_map[download_request] = len(unique_download_list)
                unique_download_list.append(download_request)
            mapping_list.append(unique_requests_map[download_request])
        return unique_download_list, mapping_list

    def _preprocess_request(self, save_data: bool, return_data: bool) -> None:
        """Prepares requests for download and creates empty folders

        :param save_data: Tells whether to save data or not
        :param return_data: Tells whether to return data or not
        """
        if not self.is_valid_request():
            raise ValueError("Cannot obtain data because request is invalid")

        if save_data:
            if self.data_folder is None:
                raise ValueError(
                    "Request parameter `data_folder` is not specified. "
                    "In order to save data please set `data_folder` to location on your disk."
                )

            for folder in self.folder_list:
                make_folder(os.path.join(self.data_folder, folder))

        for download_request in self.download_list:
            download_request.save_response = save_data
            download_request.return_data = return_data
            download_request.data_folder = self.data_folder


class FeatureIterator(Generic[_T], metaclass=ABCMeta):
    """An implementation of a base feature iteration class

    Main functionalities:

    - The iterator will load only as many features as needed at any moment
    - It will keep downloaded features in memory so that iterating over it again will not have to download the same
      features again.
    """

    def __init__(self, client: DownloadClient, url: str, params: Optional[JsonDict] = None):
        """
        :param client: An instance of a download client object
        :param url: A URL where requests will be made
        :param params: Parameters to be sent with each request
        """
        self.client = client
        self.url = url
        self.params = params or {}

        self.index = 0
        self.features: List[_T] = []
        self.finished = False

    def __iter__(self) -> "FeatureIterator[_T]":
        """Method called at the beginning of a new iteration

        :return: It returns the iterator class itself
        """
        self.index = 0
        return self

    def __next__(self) -> _T:
        """Method called to provide the next feature in iteration

        :return: the next feature
        """
        while self.index >= len(self.features) and not self.finished:
            new_features = self._fetch_features()
            self.features.extend(new_features)

        if self.index < len(self.features):
            self.index += 1
            return self.features[self.index - 1]

        raise StopIteration

    @abstractmethod
    def _fetch_features(self) -> Iterable[_T]:
        """Collects and returns more features from the service"""



"""
Module implementing geometry classes
"""
from abc import ABCMeta, abstractmethod
from math import ceil
from typing import Iterator, List, Optional, Tuple, TypeVar, Union

import shapely.geometry
import shapely.ops
import shapely.wkt
from shapely.geometry import MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry

from .constants import CRS
from .geo_utils import transform_point

Self = TypeVar("Self", bound="_BaseGeometry")


class _BaseGeometry(metaclass=ABCMeta):
    """Base geometry class"""

    def __init__(self, crs: CRS):
        """
        :param crs: Coordinate reference system of the geometry
        """
        self._crs = CRS(crs)

    @property
    def crs(self) -> CRS:
        """Returns the coordinate reference system (CRS)

        :return: Coordinate reference system Enum
        """
        return self._crs

    @property
    @abstractmethod
    def geometry(self) -> Union[Polygon, MultiPolygon]:
        """An abstract property - every subclass must implement geometry property"""

    @property
    def geojson(self) -> dict:
        """Returns representation in a GeoJSON format. Use `json.dump` for writing it to file.

        :return: A dictionary in GeoJSON format
        """
        return self.get_geojson(with_crs=True)

    def get_geojson(self, with_crs: bool = True) -> dict:
        """Returns representation in a GeoJSON format. Use `json.dump` for writing it to file.

        :param with_crs: A flag indicating if GeoJSON dictionary should contain CRS part
        :return: A dictionary in GeoJSON format
        """
        geometry_geojson = shapely.geometry.mapping(self.geometry)

        if with_crs:
            return {**self._crs_to_geojson(), **geometry_geojson}
        return geometry_geojson

    def _crs_to_geojson(self) -> dict:
        """Helper method which generates part of GeoJSON format related to CRS"""
        return {"crs": {"type": "name", "properties": {"name": f"urn:ogc:def:crs:EPSG::{self.crs.value}"}}}

    @property
    def wkt(self) -> str:
        """Transforms geometry object into `Well-known text` format

        :return: string in WKT format
        """
        return self.geometry.wkt

    @abstractmethod
    def transform(self: Self, crs: CRS, always_xy: bool = True) -> Self:
        """Transforms geometry from current CRS to target CRS."""


class BBox(_BaseGeometry):
    """Class representing a bounding box in a given CRS.

    Throughout the sentinelhub package this class serves as the canonical representation of a bounding
    box. It can initialize itself from multiple representations:

        1) ``((min_x,min_y),(max_x,max_y))``,
        2) ``(min_x,min_y,max_x,max_y)``,
        3) ``[min_x,min_y,max_x,max_y]``,
        4) ``[[min_x, min_y],[max_x,max_y]]``,
        5) ``[(min_x, min_y),(max_x,max_y)]``,
        6) ``([min_x, min_y],[max_x,max_y])``,
        7) ``'min_x,min_y,max_x,max_y'``,
        8) ``{'min_x':min_x, 'max_x':max_x, 'min_y':min_y, 'max_y':max_y}``,
        9) ``bbox``, where ``bbox`` is an instance of ``BBox``.

    Note that BBox coordinate system depends on ``crs`` parameter:

    - In case of ``constants.CRS.WGS84`` axis x represents longitude and axis y represents latitude.
    - In case of ``constants.CRS.POP_WEB`` axis x represents easting and axis y represents northing.
    - In case of ``constants.CRS.UTM_*`` axis x represents easting and axis y represents northing.
    """

    def __init__(self, bbox: Union["BBox", tuple, list, dict, str, BaseGeometry], crs: CRS):
        """
        :param bbox: A bbox in any valid representation
        :param crs: Coordinate reference system of the bounding box
        """
        x_fst, y_fst, x_snd, y_snd = BBox._to_tuple(bbox)
        self.min_x = min(x_fst, x_snd)
        self.max_x = max(x_fst, x_snd)
        self.min_y = min(y_fst, y_snd)
        self.max_y = max(y_fst, y_snd)

        super().__init__(crs)

    def __iter__(self) -> Iterator[float]:
        """This method enables iteration over coordinates of bounding box"""
        return iter(self.lower_left + self.upper_right)

    def __repr__(self) -> str:
        """Class representation"""
        return f"{self.__class__.__name__}(({self.lower_left}, {self.upper_right}), crs={repr(self.crs)})"

    def __str__(self, reverse: bool = False) -> str:
        """Transforms bounding box into a string of coordinates

        :param reverse: `True` if x and y coordinates should be switched and `False` otherwise
        :return: String of coordinates
        """
        if reverse:
            return f"{self.min_y},{self.min_x},{self.max_y},{self.max_x}"
        return f"{self.min_x},{self.min_y},{self.max_x},{self.max_y}"

    def __eq__(self, other: object) -> bool:
        """Method for comparing two bounding boxes

        :param other: Another bounding box object
        :return: `True` if bounding boxes have the same coordinates and the same CRS and `False otherwise
        """
        if isinstance(other, BBox):
            return list(self) == list(other) and self.crs is other.crs
        return False

    @property
    def lower_left(self) -> Tuple[float, float]:
        """Returns the lower left vertex of the bounding box

        :return: min_x, min_y
        """
        return self.min_x, self.min_y

    @property
    def upper_right(self) -> Tuple[float, float]:
        """Returns the upper right vertex of the bounding box

        :return: max_x, max_y
        """
        return self.max_x, self.max_y

    @property
    def middle(self) -> Tuple[float, float]:
        """Returns the middle point of the bounding box

        :return: middle point
        """
        return (self.min_x + self.max_x) / 2, (self.min_y + self.max_y) / 2

    def reverse(self) -> "BBox":
        """Returns a new BBox object where x and y coordinates are switched

        :return: New BBox object with switched coordinates
        """
        return BBox((self.min_y, self.min_x, self.max_y, self.max_x), crs=self.crs)

    def transform(self, crs: CRS, always_xy: bool = True) -> "BBox":
        """Transforms BBox from current CRS to target CRS

        This transformation will take lower left and upper right corners of the bounding box, transform these 2 points
        and define a new bounding box with them. The resulting bounding box might not completely cover the original
        bounding box but at least the transformation is reversible.

        :param crs: target CRS
        :param always_xy: Parameter that is passed to `pyproj.Transformer` object and defines axis order for
            transformation. The default value `True` is in most cases the correct one.
        :return: Bounding box in target CRS
        """
        new_crs = CRS(crs)
        return BBox(
            (
                transform_point(self.lower_left, self.crs, new_crs, always_xy=always_xy),
                transform_point(self.upper_right, self.crs, new_crs, always_xy=always_xy),
            ),
            crs=new_crs,
        )

    def transform_bounds(self, crs: CRS, always_xy: bool = True) -> "BBox":
        """Alternative way to transform BBox from current CRS to target CRS.

        This transformation will transform the bounding box geometry to another CRS as a geometric object, and then
        define a new bounding box from boundaries of that geometry. The resulting bounding box might be larger than
        original bounding box, but it will always completely cover it.

        :param crs: target CRS
        :param always_xy: Parameter that is passed to `pyproj.Transformer` object and defines axis order for
            transformation. The default value `True` is in most cases the correct one.
        :return: Bounding box in target CRS
        """
        bbox_geometry = Geometry(self.geometry, self.crs)
        bbox_geometry = bbox_geometry.transform(crs, always_xy=always_xy)
        return bbox_geometry.bbox

    def buffer(self, buffer: Union[float, Tuple[float, float]], *, relative: bool = True) -> "BBox":
        """Provides a new bounding box with a size that is changed either by a relative or an absolute buffer.

        :param buffer: The buffer can be provided either as a single number or a tuple of 2 numbers, one for buffer in
            horizontal direction and one for buffer in vertical direction. The buffer can also be negative as long as
            this doesn't reduce the bounding box into nothing.
        :param relative: If `True` the given buffer values will be interpreted as a percentage of distance between
            bounding box center point and its side edge (not to distance between opposite sides!). If `False` the given
            buffer will be interpreted as an absolute buffer measured in bounding box coordinate units.
        :return: A new bounding box of buffered size.
        """
        if isinstance(buffer, tuple):
            buffer_x, buffer_y = buffer
        elif isinstance(buffer, (int, float)):
            buffer_x, buffer_y = buffer, buffer
        else:
            raise ValueError(f"Buffer should be a number or a tuple of 2 numbers, got {type(buffer)}")

        size_x, size_y = self.max_x - self.min_x, self.max_y - self.min_y

        if relative:
            buffer_x = buffer_x * size_x / 2
            buffer_y = buffer_y * size_y / 2

        for absolute_buffer, size, direction in [(buffer_x, size_x, "horizontal"), (buffer_y, size_y, "vertical")]:
            if 2 * absolute_buffer + size <= 0:
                raise ValueError(
                    f"Negative buffer is too large, cannot reduce the bounding box to nothing in {direction} direction"
                )

        return BBox(
            (
                self.min_x - buffer_x,
                self.min_y - buffer_y,
                self.max_x + buffer_x,
                self.max_y + buffer_y,
            ),
            self.crs,
        )

    def get_polygon(self, reverse: bool = False) -> Tuple[Tuple[float, float], ...]:
        """Returns a tuple of coordinates of 5 points describing a polygon. Points are listed in clockwise order, first
        point is the same as the last.

        :param reverse: `True` if x and y coordinates should be switched and `False` otherwise
        :return: `((x_1, y_1), ... , (x_5, y_5))`
        """
        bbox = self.reverse() if reverse else self
        polygon = (
            (bbox.min_x, bbox.min_y),
            (bbox.min_x, bbox.max_y),
            (bbox.max_x, bbox.max_y),
            (bbox.max_x, bbox.min_y),
            (bbox.min_x, bbox.min_y),
        )
        return polygon

    @property
    def geometry(self) -> shapely.geometry.Polygon:
        """Returns polygon geometry in shapely format

        :return: A polygon in shapely format
        """
        return shapely.geometry.Polygon(self.get_polygon())

    def get_partition(
        self,
        num_x: Optional[int] = None,
        num_y: Optional[int] = None,
        size_x: Optional[float] = None,
        size_y: Optional[float] = None,
    ) -> List[List["BBox"]]:
        """Partitions bounding box into smaller bounding boxes of the same size.

        If `num_x` and `num_y` are specified, the total number of BBoxes is know but not the size. If `size_x` and
        `size_y` are provided, the BBox size is fixed but the number of BBoxes is not known in advance. In the latter
        case, the generated bounding boxes might cover an area larger than the parent BBox.

        :param num_x: Number of parts BBox will be horizontally divided into.
        :param num_y: Number of parts BBox will be vertically divided into.
        :param size_x: Physical dimension of BBox along easting coordinate
        :param size_y: Physical dimension of BBox along northing coordinate
        :return: Two-dimensional list of smaller bounding boxes. Their location is
        """
        if (num_x is not None and num_y is not None) and (size_x is None and size_y is None):
            size_x, size_y = (self.max_x - self.min_x) / num_x, (self.max_y - self.min_y) / num_y
        elif (size_x is not None and size_y is not None) and (num_x is None and num_y is None):
            num_x, num_y = ceil((self.max_x - self.min_x) / size_x), ceil((self.max_y - self.min_y) / size_y)
        else:
            raise ValueError("Not supported partition. Either (num_x, num_y) or (size_x, size_y) must be specified")

        return [
            [
                BBox(
                    [
                        self.min_x + i * size_x,
                        self.min_y + j * size_y,
                        self.min_x + (i + 1) * size_x,
                        self.min_y + (j + 1) * size_y,
                    ],
                    crs=self.crs,
                )
                for j in range(num_y)
            ]
            for i in range(num_x)
        ]

    def get_transform_vector(self, resx: float, resy: float) -> Tuple[float, float, float, float, float, float]:
        """Given resolution it returns a transformation vector

        :param resx: Resolution in x direction
        :param resy: Resolution in y direction
        :return: A tuple with 6 numbers representing transformation vector
        """
        return self.min_x, self._parse_resolution(resx), 0, self.max_y, 0, -self._parse_resolution(resy)

    @staticmethod
    def _parse_resolution(res: Union[str, int, float]) -> float:
        """Helper method for parsing given resolution. It will also try to parse a string into float

        :return: A float value of resolution
        """
        if isinstance(res, str):
            return float(res.strip("m"))
        if isinstance(res, (int, float)):
            return float(res)

        raise TypeError(f"Resolution should be a float, got resolution of type {type(res)}")

    @staticmethod
    def _to_tuple(bbox: Union["BBox", tuple, list, dict, str, BaseGeometry]) -> Tuple[float, float, float, float]:
        """Converts the input bbox representation (see the constructor docstring for a list of valid representations)
        into a flat tuple

        :param bbox: A bbox in one of 7 forms listed in the class description.
        :return: A flat tuple of size
        :raises: TypeError
        """
        if isinstance(bbox, (list, tuple)):
            return BBox._tuple_from_list_or_tuple(bbox)
        if isinstance(bbox, str):
            return BBox._tuple_from_str(bbox)
        if isinstance(bbox, dict):
            return BBox._tuple_from_dict(bbox)
        if isinstance(bbox, BBox):
            return BBox._tuple_from_bbox(bbox)
        if isinstance(bbox, BaseGeometry):
            return bbox.bounds
        raise TypeError("Invalid bbox representation")

    @staticmethod
    def _tuple_from_list_or_tuple(bbox: Union[list, tuple]) -> Tuple[float, float, float, float]:
        """Converts a list or tuple representation of a bbox into a flat tuple representation.

        :param bbox: a list or tuple with 4 coordinates that is either flat or nested
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: TypeError
        """
        if len(bbox) == 4:
            return tuple(map(float, bbox))  # type: ignore # iterator length
        if len(bbox) == 2 and all(isinstance(point, (list, tuple)) for point in bbox):
            return BBox._tuple_from_list_or_tuple(bbox[0] + bbox[1])
        raise TypeError("Expected a valid list or tuple representation of a bbox")

    @staticmethod
    def _tuple_from_str(bbox: str) -> Tuple[float, float, float, float]:
        """Parses a string of numbers separated by any combination of commas and spaces

        :param bbox: e.g. str of the form `min_x ,min_y  max_x, max_y`
        :return: tuple (min_x,min_y,max_x,max_y)
        """
        return tuple(float(s) for s in bbox.replace(",", " ").split() if s)  # type: ignore # iterator length

    @staticmethod
    def _tuple_from_dict(bbox: dict) -> Tuple[float, float, float, float]:
        """Converts a dictionary representation of a bbox into a flat tuple representation

        :param bbox: a dict with keys "min_x, "min_y", "max_x", and "max_y"
        :return: tuple (min_x,min_y,max_x,max_y)
        :raises: KeyError
        """
        return bbox["min_x"], bbox["min_y"], bbox["max_x"], bbox["max_y"]

    @staticmethod
    def _tuple_from_bbox(bbox: "BBox") -> Tuple[float, float, float, float]:
        """Converts a BBox instance into a tuple

        :param bbox: An instance of the BBox type
        :return: tuple (min_x, min_y, max_x, max_y)
        """
        return bbox.lower_left + bbox.upper_right


class Geometry(_BaseGeometry):
    """A class that combines shapely geometry with coordinate reference system. It currently supports polygons and
    multipolygons.

    It can be initialized with any of the following geometry representations:
    - `shapely.geometry.Polygon` or `shapely.geometry.MultiPolygon`
    - A GeoJSON dictionary with (multi)polygon coordinates
    - A WKT string with (multi)polygon coordinates
    """

    def __init__(self, geometry: Union[Polygon, MultiPolygon, dict, str], crs: CRS):
        """
        :param geometry: A polygon or multipolygon in any valid representation
        :param crs: Coordinate reference system of the geometry
        """
        self._geometry = self._parse_geometry(geometry)

        super().__init__(crs)

    def __repr__(self) -> str:
        """Method for class representation"""
        return f"{self.__class__.__name__}({self.wkt}, crs={repr(self.crs)})"

    def __eq__(self, other: object) -> bool:
        """Method for comparing two Geometry classes

        :param other: Another Geometry object
        :return: `True` if geometry objects have the same geometry and CRS and `False` otherwise
        """
        if isinstance(other, Geometry):
            return self.geometry == other.geometry and self.crs is other.crs
        return False

    def reverse(self) -> "Geometry":
        """Returns a new Geometry object where x and y coordinates are switched

        :return: New Geometry object with switched coordinates
        """
        return Geometry(shapely.ops.transform(lambda x, y: (y, x), self.geometry), crs=self.crs)

    def transform(self, crs: CRS, always_xy: bool = True) -> "Geometry":
        """Transforms Geometry from current CRS to target CRS

        :param crs: target CRS
        :param always_xy: Parameter that is passed to `pyproj.Transformer` object and defines axis order for
            transformation. The default value `True` is in most cases the correct one.
        :return: Geometry in target CRS
        """
        new_crs = CRS(crs)

        geometry = self.geometry
        if new_crs is not self.crs:
            transform_function = self.crs.get_transform_function(new_crs, always_xy=always_xy)
            geometry = shapely.ops.transform(transform_function, geometry)

        return Geometry(geometry, crs=new_crs)

    @classmethod
    def from_geojson(cls, geojson: dict, crs: Optional[CRS] = None) -> "Geometry":
        """Create Geometry object from geojson. It will parse crs from geojson (if info is available),
        otherwise it will be set to crs (WGS84 if parameter is empty)

        :param geojson: geojson geometry (single feature)
        :param crs: crs to be used if not available in geojson, CRS.WGS84 if not provided
        :return: Geometry object
        """
        try:
            crs = CRS(geojson["crs"]["properties"]["name"])
        except (KeyError, AttributeError, TypeError):
            pass

        if not crs:
            crs = CRS.WGS84

        return cls(geojson, crs=crs)

    @property
    def geometry(self) -> Union[Polygon, MultiPolygon]:
        """Returns shapely object representing geometry in this class

        :return: A polygon or a multipolygon in shapely format
        """
        return self._geometry

    @property
    def bbox(self) -> BBox:
        """Returns BBox object representing bounding box around the geometry

        :return: A bounding box, with same CRS
        """
        return BBox(self.geometry, self.crs)

    @staticmethod
    def _parse_geometry(geometry: Union[Polygon, MultiPolygon, dict, str]) -> Union[Polygon, MultiPolygon]:
        """Parses given geometry into shapely object

        :param geometry: A representation of the geometry
        :return: Shapely polygon or multipolygon
        :raises TypeError
        """
        if isinstance(geometry, str):
            geometry = shapely.wkt.loads(geometry)
        elif isinstance(geometry, dict):
            geometry = shapely.geometry.shape(geometry)
        elif not isinstance(geometry, BaseGeometry):
            raise TypeError("Unsupported geometry representation")

        if not isinstance(geometry, (Polygon, MultiPolygon)):
            raise ValueError(f"Supported geometry types are polygon and multipolygon, got {type(geometry)}")

        return geometry


class BBoxCollection(_BaseGeometry):
    """A collection of bounding boxes"""

    def __init__(self, bbox_list: Union["BBoxCollection", List[BBox]]):
        """
        :param bbox_list: A list of BBox objects which have to be in the same CRS
        """
        self._bbox_list, crs = self._parse_bbox_list(bbox_list)
        self._geometry = self._get_geometry()

        super().__init__(crs)

    def __repr__(self) -> str:
        """Method for class representation"""
        bbox_list_repr = ", ".join([repr(bbox) for bbox in self.bbox_list])
        return f"{self.__class__.__name__}({bbox_list_repr})"

    def __eq__(self, other: object) -> bool:
        """Method for comparing two BBoxCollection classes"""
        if isinstance(other, BBoxCollection):
            return (
                self.crs is other.crs
                and len(self.bbox_list) == len(other.bbox_list)
                and all(bbox == other_bbox for bbox, other_bbox in zip(self, other))
            )
        return False

    def __iter__(self) -> Iterator[BBox]:
        """This method enables iteration over bounding boxes in collection"""
        return iter(self.bbox_list)

    @property
    def bbox_list(self) -> List[BBox]:
        """Returns the list of bounding boxes from collection

        :return: The list of bounding boxes
        """
        return self._bbox_list

    @property
    def geometry(self) -> MultiPolygon:
        """Returns shapely object representing geometry

        :return: A multipolygon of bounding boxes
        """
        return self._geometry

    @property
    def bbox(self) -> BBox:
        """Returns BBox object representing bounding box around the geometry

        :return: A bounding box, with same CRS
        """
        return BBox(self.geometry, self.crs)

    def reverse(self) -> "BBoxCollection":
        """Returns a new BBoxCollection object where all x and y coordinates are switched

        :return: New Geometry object with switched coordinates
        """
        return BBoxCollection([bbox.reverse() for bbox in self.bbox_list])

    def transform(self, crs: CRS, always_xy: bool = True) -> "BBoxCollection":
        """Transforms BBoxCollection from current CRS to target CRS

        :param crs: target CRS
        :param always_xy: Parameter that is passed to `pyproj.Transformer` object and defines axis order for
            transformation. The default value `True` is in most cases the correct one.
        :return: BBoxCollection in target CRS
        """
        return BBoxCollection([bbox.transform(crs, always_xy=always_xy) for bbox in self.bbox_list])

    def _get_geometry(self) -> MultiPolygon:
        """Creates a multipolygon of bounding box polygons"""
        return MultiPolygon([bbox.geometry for bbox in self.bbox_list])

    @staticmethod
    def _parse_bbox_list(bbox_list: Union["BBoxCollection", List[BBox]]) -> Tuple[List[BBox], CRS]:
        """Helper method for parsing a list of bounding boxes"""
        if isinstance(bbox_list, BBoxCollection):
            return bbox_list.bbox_list, bbox_list.crs

        if not isinstance(bbox_list, list) or not bbox_list:
            raise ValueError("Expected non-empty list of BBox objects")

        for bbox in bbox_list:
            if not isinstance(bbox, BBox):
                raise ValueError(f"Elements in the list should be of type {BBox.__name__}, got {type(bbox)}")

        crs = bbox_list[0].crs
        for bbox in bbox_list:
            if bbox.crs is not crs:
                raise ValueError("All bounding boxes should have the same CRS")

        return bbox_list, crs
